/** TDSP *******************************************************************
                          TDSPLTVChannel.h  -  description
                             -------------------
    begin                : Don Jul 17 2003
    copyright            : (C) 2003 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPLTVCHANNEL_H
#define TDSPLTVCHANNEL_H

#include <TDSPOutput1.h>
#include <TDSPBlockOperation.h>

/**This is the base class for Linear Time-Variant Channels
 *@author Peter Jung
 */

class TDSPLTVChannel : public TDSPOutput1, public TDSPBlockOperation  {

 private:
  // This class already implements some very basic channels 
  // these variables are need for this channels 

  TDSPVector* fPowerDelayProfileSqrt; ///< Simple (Sqrt of the) Power Delay Profile for rayleigh/fixed etc. channels
  UInt_t      fBasicChannelType; ///< Basic Channels (0=not used,1=fixed LTI-filter, 2=RandNC, 3=Rayleigh(no phase)
  UInt_t      fDebugBits; ///< Several Bits that can set for debugging

 protected:

  Int_t       fBlockLen; ///< Block processing
  TDSPMatrix* fImpulseResponse; ///< time-variant impulse response h(tau,t) (t=columns, tau=delay)
  TDSPMatrix* fChannelMatrix; ///< momentan channel matrix
  Int_t       fChannelMatrixSlot; ///< the time-slot which was used in the "fChannelMatrix"-calculation
  Int_t       fChanMatSize; ///< size of channel matrices to calculate(depends on the outer decisions...)


  Bool_t   fHaveRepresentation; ///< ChangeRepresentation already called ?
  Bool_t   fHaveImpulseResponse; ///< fImpulseResponse uptodate ?
  Bool_t   fHaveChannelMatrix; ///< fChannelMatrix uptodate ?

  Bool_t   fCalculateChannelMatrix; ///< Do have to calculate and emit the channel matrices ?
  Bool_t   fCalculateChannelMatrixCyclic; ///< Cyclic Channel Matrix ?


 public: 
  TDSPLTVChannel(char *name=NULL);
  ~TDSPLTVChannel();

  virtual void        Print();
  virtual Int_t       GetBlockLen() { return fBlockLen;};
  virtual void        SetBlockLen(Int_t bl) { fBlockLen=bl;};

  // Set Some Basic Channel Models
  // mod = "none" 
  // mod = "fixed"  - fixed time-invariant impulse response with (sqrt-)power given by "profile"
  // mod = "randnc" - impulse response with complex gaussian entries with (sqrt-)power given by "profile"

  void SetModel(Option_t *mod, TDSPVector *profileSqrt=NULL);

  // Setting Options 
  //
  // channelmatrix   - calculate and emit channel matrices 
  // nochannelmatrix - dont calculate channel matrices
  // debugchannelmatrix - print out each channel matrix before emission
  
  void                SetOption(Option_t *opt); 

  void                SetChannelMatrixSize(Int_t L) { fChanMatSize=L;}; ///< set the matrix size of channel matrix calculations
  Int_t               GetChannelMatrixSize() const { return fChanMatSize;}; ///< get the matrix size of channel matrices ...
  
  TDSPMatrix*         GetImpulseResponse(); ///< returns h(t,tau)
  TDSPMatrix*         GetChannelMatrix(Int_t TimeSlot); ///< returns the channel matrix (in time domain) in the "TimeSlot"

  virtual TDSPMatrix* CreateImpulseResponse(TDSPMatrix *out=NULL); ///< creates h(t,tau) 
  TDSPMatrix*         CreateChannelMatrix(TDSPMatrix *out=NULL,
					 Int_t TimeStart=0,
					 Int_t TimeEnd=0); ///< creates the channel matrix from h(t,tau) in [TimeStart:TimeEnd-1]

  virtual void        ChangeRepresentation(); ///< Force a change of the channel representation

  /* Signals */

  virtual void        ImpulseResponse(TDSPMatrix*i); ///< Emit a new Impulse response
  virtual void        ChannelMatrix(TDSPMatrix*i);  ///< Emit a new Channel Matrix

  virtual TDSPSignal* Update(TDSPSignal* input);
    
  virtual void        ChannelMatrices(Int_t ilen=0); ///< Process the channel and emit the channel matrices for data of length "ilen"
  virtual TDSPVector *Apply(TDSPVector*in,TDSPVector*out = NULL);

  ClassDef(TDSPLTVChannel,1)
};

#endif
