/** TDSP *******************************************************************
                       TDSPMultiEcho.h  -  description
                             -------------------
    begin                : Die Dez 10 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPMULTIECHO_H
#define TDSPMULTIECHO_H

#include <TList.h>
#include <TGraph.h>
#include <TMultiGraph.h>
#include <TH1.h>
#include <TArrayD.h>
#include <TComplex.h>
#include <TDSPMatrix.h>
#include <TDSPLTVChannel.h>

/**
  *@author Peter Jung
  */


class TDSPSingleEcho : public TNamed {

 protected:
  Double_t fMinTau;  ///< minimum of this cluster's delay times [us]
  Double_t fTauRange;///< range   of this cluster's delay times [us]
  Double_t fFraction;  ///< EnergyFraction of in this cluster
  Double_t fTau;     ///< moment. representation of the echo arrival
  Double_t fPhase;   ///< moment. representation of the phase
  Double_t fDoppler; ///< moment. representation of the Doppler

 public:

  TDSPSingleEcho(Double_t MinTau=0, Double_t TauRange=1, Double_t EnergyFraction=1.) : 
    fMinTau(MinTau), fTauRange(TauRange), fFraction(EnergyFraction), 
    fPhase(0), fDoppler(0) {};

  ~TDSPSingleEcho() {};

  Double_t          GetPhase() const { return fPhase;};
  Double_t          GetDoppler() const { return fDoppler;};
  Double_t          GetTau() const { return fTau;};
  Double_t          GetMinTau() const { return fMinTau;};
  Double_t          GetMaxTau() const { return fTauRange+fMinTau;};
  Double_t          GetTauRange() const { return fTauRange;};
  Double_t          GetFraction() const { return fFraction;};
  virtual Double_t  GetProbability(Double_t tau);
  virtual void      ChangeRepresentation();
  TGraph*           Graph(TGraph *in=NULL, Double_t x0=-9999, Double_t x1=-9999, UInt_t num=100);
  void              Draw(Option_t *o="al", Double_t x0=-9999, Double_t x1=-9999, UInt_t num=100);
  virtual void      Print();         

};

class TDSPSingleEchoEXP : public TDSPSingleEcho {

 protected:
  Double_t fDecay; ///< exp. decay factor for the echo delay times,
 public:
  TDSPSingleEchoEXP(Double_t MinTau=0, Double_t TauRange=1, Double_t EnergyFraction=1.,
		    Double_t Decay=1.) : TDSPSingleEcho(MinTau,TauRange,EnergyFraction), fDecay(Decay) {};
  ~TDSPSingleEchoEXP() {};

  virtual void      ChangeRepresentation();
  virtual Double_t  GetProbability(Double_t tau);
  virtual void      Print(); 
  
};

class TDSPMultiEcho : public TDSPLTVChannel   {
 protected:
  
  UInt_t   fNumClusters; ///< Number of clusters to manage
  UInt_t   fModelGroup; ///< Modelgroups
  UInt_t   fModel;  ///< Model
  UInt_t   fNumEchos; ///< number of Echos
  Double_t fSymbolRate; ///< Symbol sampling rate
  UInt_t   fOverSampling; ///< Oversampling Factor
  Double_t fMaxDoppler; ///< Maximum Doppler frequency  fDmax := v*f0/c0
  UInt_t   fNumSlices; ///< Number of points in the time to determine the impulse response
  Double_t fCarrierFreq; ///< Carrier Frequency in Hz

  TList*   fListOfEchos; ///< the List of the Single Echos
  TArrayD* fTaus; ///< The mom. representation of all echo arrivals
  TArrayI* fDiscreteTaus; ///< The mom. representation of all quantisized echo arrivals
  TArrayD* fPhase; ///< The mom. representation of all phases
  TArrayD* fDoppler;  ///< The mom. representation of all doppler frequencies

  TDSPVector* fFilter; ///< band limiting filter
  TDSPMatrix* fImpulseResponse; ///< time-variant impulse response h(t,tau)
  TDSPMatrix* fTransferFunction; ///< the time-variant transfer function h(t,f)
  TDSPMatrix* fScatteringFunction; ///< the scattering function C_H(tau,nu)

  void Add(TDSPSingleEcho* e) { fListOfEchos->Add(e);fNumClusters=fListOfEchos->GetSize();};
  void SetModel(UInt_t ModelGroup, UInt_t Model); ///< Initialize a model
  void CleanModel(); ///< Clean the model
  
  Bool_t   fHaveRepresentation; ///< ChangeRepresentation already called ?
  Bool_t   fHaveImpulseResponse; ///< fImpulseResponse uptodate ?
  Bool_t   fHaveTransferFunction; ///< fTransferFunction uptodate ?
  Bool_t   fRandomPhases; ///< Use the random phases of the echos ?

 public: 
  TDSPMultiEcho(char *name="",char *opt="", UInt_t numEcho=50, Double_t symRate=1.0e6, 
		Double_t maxdoppler=100./3.6/3.0e8*5200.*1.0e6 /*=100kmh at 5.2 GHz*/);
  ~TDSPMultiEcho();
  
  virtual void       SetBlockLen(Int_t bl) { SetNumSlices(bl);};
  virtual Int_t      GetBlockLen() { return GetNumSlices();};

  virtual void       Print();
  virtual void       ls();
  void               SetOption(Option_t *o);
  void               SetNumEchos(UInt_t N);
  void               SetNumSlices(UInt_t N) { fNumSlices=N;};
  void               SetSymbolRate(Double_t s) { fSymbolRate=s;};
  void               SetSymbolRateMHz(Double_t s) { SetSymbolRate(s*1.0e6);};
  void               SetOverSampling(UInt_t s) { fOverSampling=s;};
  void               SetMaxDoppler(Double_t d) { fMaxDoppler=d;};
  void               SetSpeed(Double_t v) { SetMaxDoppler(v/3.0e8*fCarrierFreq);}; ///< Set Speed in m/s
  void               SetSpeedKMH(Double_t v) { SetSpeed(v*1000./3600.);}; ///< Set Speed in km/h
  void               SetCarrierFrequency(Double_t f) { fCarrierFreq=f;}; ///< Set the carrier frequency in Hz
  void               SetCarrierFrequencyMHz(Double_t f) { SetCarrierFrequency(f*1.0e6); }; ///< Set the carrier frequency in MHz
  void               SetCarrierFrequencyGHz(Double_t f) { SetCarrierFrequency(f*1.0e9); }; ///< Set the carrier frequency in GHz
  
  virtual void       ChangeRepresentation();
  TArrayD*           GetTaus() const { return fTaus;};

  virtual Double_t   GetMinTau();
  virtual Double_t   GetMaxTau();
  virtual Double_t   GetTauRange();
  virtual Double_t   GetMinTauRange();
  virtual Double_t   GetMaxTauRange();

  UInt_t             GetNumEchos() const { return fNumEchos;};
  Double_t           GetSymbolRate() const { return fSymbolRate;};
  Double_t           GetSymbolRateMHz() { return GetSymbolRate()/1.0e6;};
  Double_t           GetSampleRate() { return GetSymbolRate()*fOverSampling;};
  Double_t           GetSampleRateMHz() { return GetSymbolRateMHz()*fOverSampling;};
  UInt_t             GetOverSampling() const { return fOverSampling;};
  Double_t           GetMaxDoppler() const { return fMaxDoppler;}
  UInt_t             GetNumSlices() const { return fNumSlices;};
  Double_t           GetSpeed() { return fMaxDoppler*3.0e8/fCarrierFreq;};
  Double_t           GetSpeedKMH() { return GetSpeed()/1000.*3600.;};
  Double_t           GetCarrierFrequency() const { return fCarrierFreq;};
  Double_t           GetCarrierFrequencyMHz() { return GetCarrierFrequency()/1.0e6;};
  Double_t           GetCarrierFrequencyGHz() { return GetCarrierFrequency()/1.0e9;};

  virtual Double_t   GetProbability(Double_t tau);
  TGraph*            Graph(TGraph *in=NULL, Double_t x0=-9999, Double_t x1=-9999, UInt_t num=100);
  TMultiGraph*       MultiGraph(TMultiGraph *in=NULL,Double_t x0=-9999, Double_t x1=-9999, UInt_t num=100);
  void               Draw(Option_t *o="al", Double_t x0=-9999, Double_t x1=-9999, UInt_t num=100);
  TDSPMatrix*        GetTransferFunction(); ///< returns h(t,f)
  TDSPMatrix*        GetScatteringFunction(); ///< return C_H(tau,nu)

  TDSPMatrix*        CreateImpulseResponse(TDSPMatrix *out=NULL); ///< creates h(t,tau) 
  TDSPMatrix*        CreateTransferFunction(TDSPMatrix *out=NULL); ///< creates h(t,f)
  TDSPMatrix*        CreateScatteringFunction(TDSPMatrix *out=NULL); ///< creates C_H(tau,nu)
  TH1F*              Fill(UInt_t num=1, TH1F* h=NULL); ///< fills num representation into a histogram

  ClassDef(TDSPMultiEcho,1)

};

#endif
