/** TDSP *******************************************************************
                          TDSPWssus.h  -  description
                             -------------------
    begin                : Don Jun 27 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPWSSUS_H
#define TDSPWSSUS_H

#include <TDSPVector.h>
#include <TDSPMatrix.h>
#include <TDSPLTVChannel.h>

/**implementation of the WSSUS-Channel
 *@author Peter Jung
 */

class TDSPWssus : public TDSPLTVChannel  {
 protected:

  Int_t       fType; ///< Scattering function type (1=jakes, 2=flat) (use SetOption(..))

  TDSPMatrix* fTransferFunction; ///< the time-variant transfer function h(t,f)
  TDSPMatrix* fScatteringFunction; ///< the scattering function C_H(tau,nu)
  TDSPMatrix* fSpreadingFunction; ///< momentan scattering

  Double_t    fMaxDoppler;    ///< Maximum Doppler frequency  fDmax := v*f0/c0
  Double_t    fCarrierFreq;   ///< Carrier Frequency in Hz
  Double_t    fSymbolRate; ///< Symbol sampling rate (sampling rate is  fSymbolRate*fOverSampling
  Double_t    fOverSampling; ///< Oversampling Factor

  Double_t    fExpDecay; ///< Decay 
  Double_t    fMaxDelay; ///< Maximum Delay time


  static TDSPMatrix* 
    ScatFunction(Int_t     type,
		 Double_t  alpha,
		 Double_t  beta,
		 UInt_t    t0,
		 UInt_t    f0,
		 UInt_t    clen,
		 TDSPMatrix *ScatFun=NULL);

  static TDSPMatrix*
    TapFunction(TDSPMatrix *ScatFun,       // The Scattering Matrix
		UInt_t      tlen,          // Maximum Time Delay
		UInt_t      flen,          // Maximum Doppler Shift
		UInt_t      TapFunLen,     // The length of the tapfunction
		TDSPMatrix *TapFun=NULL);  // The tapfun itself
  
 public: 
  TDSPWssus(char *name = NULL);
  ~TDSPWssus();
  
  // Setting Options
  //
  // jakes - jakes spectrum
  // flat  - flat spectrum
  // or - see more : TDSPLTVChannel::SetOption(..)

  void               SetOption(Option_t *opt); 
  
  void               SetSymbolRate(Double_t s) { fSymbolRate=s;};
  void               SetSymbolRateMHz(Double_t s) { SetSymbolRate(s*1.0e6);};
  void               SetOverSampling(Double_t s) { fOverSampling=s;};
  void               SetMaxDoppler(Double_t d) { fMaxDoppler=d;};
  void               SetSpeed(Double_t v) { SetMaxDoppler(v/3.0e8*fCarrierFreq);}; ///< Set Speed in m/s
  void               SetSpeedKMH(Double_t v) { SetSpeed(v*1000./3600.);}; ///< Set Speed in km/h
  void               SetCarrierFrequency(Double_t f) { fCarrierFreq=f;}; ///< Set the carrier frequency in Hz
  void               SetCarrierFrequencyMHz(Double_t f) { SetCarrierFrequency(f*1.0e6); }; ///< Set the carrier frequency in MHz
  void               SetCarrierFrequencyGHz(Double_t f) { SetCarrierFrequency(f*1.0e9); }; ///< Set the carrier frequency in GHz
  void               SetExpDecay(Double_t al) { fExpDecay = al;}; ///< set the exp. decay for time in seconds (Hz)
  void               SetExpDecayMHz(Double_t al) { SetExpDecay(al*1.0e6);}; ///< set the exp. decay for time in 1.e6 seconds (MHz)
  void               SetMaxDelay(Double_t md) { fMaxDelay = md;}; ///< set the maximum delay time in sec
  void               SetMaxDelayMus(Double_t md) { SetMaxDelay(md/1.0e6);}; ///< set the maximum delay time in sec

  Double_t           GetSamplePeriod() { return 1./GetSamplingRate();}; ///< return sample period in seconds
  Double_t           GetSamplePeriodMus() { return 1./GetSamplingRateMHz();}; ///< return sample period in s
  Double_t           GetSymbolPeriod() { return 1./GetSymbolRate();}; ///< return symbol period in seconds
  Double_t           GetSymbolPeriodMus() { return 1./GetSymbolRateMHz();}; ///< return symbol period in s
  Double_t           GetMaxDelay() const { return fMaxDelay;};///< maximum delay in seconds
  Double_t           GetMaxDelayMus() { return fMaxDelay*1.0e6;} ///< maximum delay in sec
  Double_t           GetExpDecay() const { return fExpDecay;}; ///< in Hz
  Double_t           GetExpDecayMHz() const { return fExpDecay/1.0e6;}; ///< in MHz
  Double_t           GetSymbolRate() const { return fSymbolRate;};
  Double_t           GetSymbolRateMHz() { return GetSymbolRate()/1.0e6;};
  Double_t           GetSamplingRate() { return GetSymbolRate()*fOverSampling;};
  Double_t           GetSamplingRateMHz() { return GetSymbolRateMHz()*fOverSampling;};
  Double_t           GetOverSampling() const { return fOverSampling;};
  Double_t           GetMaxDoppler() const { return fMaxDoppler;};
  Double_t           GetSpeed() { return fMaxDoppler*3.0e8/fCarrierFreq;};
  Double_t           GetSpeedKMH() { return GetSpeed()/1000.*3600.;};
  Double_t           GetCarrierFrequency() const { return fCarrierFreq;};
  Double_t           GetCarrierFrequencyMHz() { return GetCarrierFrequency()/1.0e6;};
  Double_t           GetCarrierFrequencyGHz() { return GetCarrierFrequency()/1.0e9;};
  Int_t              GuessChannelBlockLen(); ///< Guess a Channel Block length
  // Discrete Variables 

  Int_t    GetDiscreteMaxDelay()   {return TMath::Nint(TMath::Max(GetMaxDelay()*GetSamplingRate(),1.));};
  Double_t GetDiscreteExpDecay()   {return GetExpDecay()/GetSamplingRate();};
  Int_t    GetDiscreteMaxDoppler() {return TMath::Nint(TMath::Max(GetMaxDoppler(),1.));};
  Int_t    GetDiscreteScatteringMatrixSize(); 
  

  TDSPMatrix*        GetTransferFunction(); ///< returns h(t,f)
  TDSPMatrix*        GetScatteringFunction(); ///< return C_H(tau,nu)
  TDSPMatrix*        GetSpreadingFunction();  ///< return the spreading function h(nu,tau)
  TDSPMatrix*        CreateTransferFunction(TDSPMatrix *out=NULL); ///< creates h(t,f)
  TDSPMatrix*        CreateScatteringFunction(TDSPMatrix *out=NULL); ///< creates C_H(tau,nu)
  TDSPMatrix*        CreateSpreadingFunction(TDSPMatrix *out=NULL); ///< creates the spreading function
  virtual TDSPMatrix*CreateImpulseResponse(TDSPMatrix *out=NULL); ///< creates h(t,tau) 

  Bool_t   fHaveScatteringFunction; ///< fScatteringFunction uptodate ?
  Bool_t   fHaveTransferFunction; ///< fTransferFunction uptodate ?
  Bool_t   fHaveSpreadingFunction; ///< fSpreadingFunction uptodate ?


  

  virtual void        ChangeRepresentation(); ///< Force a change of the channel representation

  

  virtual void        Draw(Option_t *o="");
  virtual void        Print();
  
  ClassDef(TDSPWssus,1)

};

#endif
