/** TDSP *******************************************************************
                          TDSPFilterBank.h  -  description
                             -------------------
    begin                : Mon Okt 14 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPFILTERBANK_H
#define TDSPFILTERBANK_H

#include <TNamed.h>
#include <TH2.h>
#include <TDSPRating.h>
#include <TDSPVector.h>
#include <TDSPMatrix.h>

/**Filterbank implementation
 *@author Peter Jung
 */

#define OPT_MY_FILTERMATRIX 0x1
#define OPT_DFT_FILTERBANK  0x2

class TDSPFilterBank : public TNamed, virtual public TDSPRating  {

 private:

  Bool_t  _isMyMatrix() { return fOpt&OPT_MY_FILTERMATRIX;};
  void    _enableMyMatrix() { fOpt|=OPT_MY_FILTERMATRIX;};
  void    _disableMyMatrix() { fOpt&=(!OPT_MY_FILTERMATRIX);};


  Bool_t  _isDFT() { return fOpt&OPT_DFT_FILTERBANK;};
  void    _enableDFT() { fOpt|=OPT_DFT_FILTERBANK;};
  void    _disableDFT() { fOpt&=(!OPT_DFT_FILTERBANK);};

  Bool_t  _IsReady(); ///< Check that the filterbank is ready to use !
  void    _CleanMe(); ///< Clean all blocks etc.
  void    _MinMaxSubSampling(); ///< Update min/max-subsamplings

  void    _reset_options(); ///< reset the activation for each filter to "active"
  Int_t    fZero; ///< The Position of the zero in the filters
  UInt_t   fNumActive; ///< counts here the number of active filters

 protected:

  UInt_t      fOpt; ///< several options
  TDSPMatrix* fFilterBank; ///< The filters of the filterbank in matrix form
  TArrayI     fSubSamplings; ///< The Subsampling factors (=timepads)
  UInt_t      fMinSubSampling; ///< Minimal Subsampling
  UInt_t      fMaxSubSampling; ///< Maximal Subsampling
  TArrayI     fFilterOptions; ///< Several Options per Filter 
  TString     fMethodStr; ///< a method string describing the method of filter allocation

  // members used for DFT-Filterbank

  UInt_t      fFreqPad; ///< The discrete-time frequency pad

 public: 
  TDSPFilterBank(UInt_t N=0, UInt_t L=0); ///< Init with N filters of common length L
  ~TDSPFilterBank();

  void           SetFilter(UInt_t n, TDSPVector *f); ///< Set the response for filter n from a vector f
  TDSPVector*    GetFilter(UInt_t n, TDSPVector *f=NULL); ///< Return the response of filter n as a vector

  void           SetMatrix(TDSPMatrix *a); ///< set the filter matrix
  TDSPMatrix*    GetMatrix() const { return fFilterBank;}; ///< return the filter matrix
  void           LoadMatrix(char *fname, char *key); ///< Load the matrix from file "fname" under the key "key"

  void           SetSize(UInt_t N, UInt_t L); ///< Set the number of filters and the common length
  void           SetSubSampling(UInt_t n, UInt_t T); ///< Set the subsampling for filter n
  void           SetSubSamplings(UInt_t T); ///< Set the subsampling for all filters
  void           SetTimePad(UInt_t T) { SetSubSamplings(T);}; ///< Set the subsampling for all filters
  void           SetFilterOption(UInt_t opt); ///< Set all options for all filters
  void           SetFilterOption(UInt_t n, UInt_t opt); ///< Set all options for that filter
  UInt_t         GetFilterOption(UInt_t n); ///< Get all options for that filter
  Bool_t         IsActive(UInt_t n); ///< Is this filter activated
  void           Activate(UInt_t n); ///< Activate this filter
  void           Deactivate(UInt_t n); ///< Deactivate this filter

  Bool_t         IsCommonSubSampling(); ///< Is the Subsampling for all filters the same ?
  UInt_t         GetSubSampling(UInt_t n); ///< Get the subsampling for filter n
  UInt_t         GetSubSamplings(); ///< Get the subsamplings if their are equal for all filters !!
  UInt_t         GetMaxSubSampling() const { return fMaxSubSampling;}; ///< Get the largest Subsampling factor
  UInt_t         GetMinSubSampling() const { return fMinSubSampling;}; ///< Get the lowest Subsampling factor

  UInt_t         GetLength(); ///< Return the common length of filter response
  UInt_t         GetNum(); ///< Return the number of filters in the bank
  void           SetLength(UInt_t); ///< Set the length of the filters
  void           SetNum(UInt_t); ///< Set the number of the filters
  void           SetZero(Int_t z) { fZero=z;}; ///< Set the index of the zero-point in the filters
  Int_t          GetZero() const { return fZero;}; ///< Get the index of the zero-point in the filters


  TDSPVector*    Synthesize(TDSPVector *input, TDSPVector *output=NULL);  ///< Perform the synthesis operation
  TDSPVector*    Analyze( TDSPVector *input, TDSPVector *output=NULL);  ///< Perform the analysis operation

  // Automatic filters
  //
  virtual void   InitFromProtoType(Option_t* method, TDSPVector *input); ///< Init the filters by a method and a prototyp
  virtual void   LoadAndInitFromProtoType(Option_t* method,char *fname,char*key=NULL); ///< Load "fname" and Init the filters by a method and a prototyp

  // methods used for DFT-Filterbank
  //
  UInt_t         GetFreqPad() const { return fFreqPad;}; ///< Get the frequency pad (only for DFT-Filterbanks)
  void           SetFreqPad(UInt_t f) { fFreqPad=f;};///< Set the frequency pad (only for DFT-Filterbanks)
  UInt_t         GetNumActiveFilters() const { return fNumActive;}; ///< Return the number of active filters
  

  void           Print(); ///< Print out the settings
  //TH2*           Draw(Option_t *option = "hist surf4"); ///< Draw the filtermatrix 


  Double_t       GetBandWidthEfficiency();

  // Calculate the Dual Filterbank specified by a method

  TDSPMatrix*    GetDualMatrix(Option_t *method="pseudoinverse");

  ClassDef(TDSPFilterBank,1)

};

#endif
