/** TDSP *******************************************************************
                          TDSPLmmse.h  -  description
                             -------------------
    begin                : Mon Apr 29 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPLMMSE_H
#define TDSPLMMSE_H

#include <TDSPEstimator.h>

/**Linear Minimum Mean Squared Estimator
 *@author Peter Jung

 y=f(x) - finds an estimator x1 (a matrix q) with x1=qy so that E[(x1-x)^2] -> min
 
 (a) for linear case y=Hx+n it is in general q = E[xx^* H^*] E[Hxx^* H^* + nn^*]^-1
     ---> q = E[xx^*] H^* E[H*xx^* H^* + nn^*]^-1
     with Cx=E[xx^*] and Cn=E[nn^*]
     ---> q = Cx H^* (H Cx H^* + Cn)^-1
 */

class TDSPLmmse : public TDSPEstimator  {
 
  TDSPMatrix* fEstimatorMatrix;

  // for a linear equation y=Hx+n

  TComplex    fNoiseScaler; // a Scaler to the Noise Covariance-Matrix scale * E[nn*]
  TDSPMatrix *fNoiseCovariance; // The Noise Covariance-Matrix E[nn*]
  TDSPMatrix *fXCovariance; // The Covariance-Matrix of the input data E[xx*] 
  TDSPMatrix *fH; // The Matrix H in y=Hx+n

  TDSPMatrix *fHCxHStar; // this H Cx H^*
  TDSPMatrix *fPost; // Postmultiplication (from right) ...

  Bool_t _recalc1,_recalc2; // several recalculation steps
  Bool_t _my_x_matrix, _my_noise_matrix; // internal or external matrices ??

  void _init();
  void recalc1() { _recalc1=kTRUE;};
  void recalc2() { _recalc2=kTRUE;};
  
  void default_noise_covariance(Int_t rows, Int_t cols); // Set default noise covariance (unit matrix)
  void default_x_covariance(Int_t rows, Int_t cols); // Set default x covariance (unit matrix)
  void updatefHCxHStar();

 public: 
  TDSPLmmse();
  TDSPLmmse(TDSPMatrix* H); 
  TDSPLmmse(TDSPMatrix* H, TDSPMatrix* Cx, TDSPMatrix* Cn); 
    
  ~TDSPLmmse();

  TDSPVector* Estimate(TDSPVector* y);  // Estimate 
  void        Calculate();

  void        SetMatrix(TDSPMatrix*h) { fH = h;recalc1();recalc2();};
  TDSPMatrix* GetMatrix() const { return fH;};
  void        SetPostMatrix(TDSPMatrix *h) {fPost = h;};
  TDSPMatrix* GetPostMatrix() const { return fPost;};
  void        SetXCovariance(TDSPMatrix*c);
  TDSPMatrix* GetXCovariance() const { return fXCovariance;};
  void        SetNoiseCovariance(TDSPMatrix*c); 
  TDSPMatrix* GetNoiseCovariance() const { return fNoiseCovariance;};
  void        SetNoise(TComplex sigma2) { fNoiseScaler=sigma2;recalc2();};
  TComplex    GetNoise() const { return fNoiseScaler;};
  
  TDSPMatrix* GetEstimatorMatrix() { Calculate();return fEstimatorMatrix;};

  ClassDef(TDSPLmmse,1)

};

inline TDSPVector* TDSPLmmse::Estimate(TDSPVector* y) {
  Calculate();
  if (fEstimatorMatrix) {
    fEstimatorMatrix->RightApply(y,fEstimate);
  } else {
    Error("Estimate","No Estimation Matrix specified !!");
    return NULL;
  }
  return fEstimate;
}

#endif
