/** TDSP *******************************************************************
                          TDSPMath.h  -  description
                             -------------------
    begin                : Thu Jan 31 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPMATH_H
#define TDSPMATH_H


#include <TMath.h>

/**Additional Mathematical Functions
   specialized to signal processing
   *@author Peter Jung
   */

class TDSPMath {
 public: 
  
  static  Double_t  BesselJ0(Double_t x);
  static  Double_t  Q(Double_t x);
  static  Double_t  QSqrt(Double_t x);
  static  Double_t  Rayleigh(Double_t x,Double_t sigma = 1);
  static  Double_t  dB(Double_t x);
  static  Double_t  FromdB(Double_t x);
  static  Double_t  Lorentz(Double_t x, Double_t sigma = 1);

  ClassDef(TDSPMath,0)

};


// Because BesselJ is not yet in the root-TMath-Lib 
//
// This i adapted from : http://www.if.ufrj.br/teaching/compute/cpp/Bessel_functions.cpp
// 

inline Double_t TDSPMath::BesselJ0(Double_t x) {
  Double_t ax,z;
  Double_t xx,y,ans,ans1,ans2;		/* Acumulate polynomials in Double_t */
  
  if ((ax=TMath::Abs(x)) < 8.0) {		/* Direct rational function fit. */
    y=x*x;
    ans1=57568490574.0+y*(-13362590354.0+y*(651619640.7
					    +y*(-11214424.18+y*(77392.33017+y*(-184.9052456)))));
    ans2=57568490411.0+y*(1029532985.0+y*(9494680.718
					  +y*(59272.64853+y*(267.8532712+y*1.0))));
    ans=ans1/ans2;
  } else {		/* Fitting function. */
    z=8.0/ax;
    y=z*z;
    xx=ax-0.785398164;
    ans1=1.0+y*(-0.1098628627e-2+y*(0.2734510407e-4
				    +y*(-0.2073370639e-5+y*0.2093887211e-6)));
    ans2 = -0.1562499995e-1+y*(0.1430488765e-3
			       +y*(-0.6911147651e-5+y*(0.7621095161e-6
						       -y*0.934935152e-7)));
    ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2);
  }
  return ans;
}                           

inline Double_t TDSPMath::Q(Double_t x) {
  return 0.5*TMath::Erfc(x/TMath::Sqrt(2.0));
}

inline Double_t TDSPMath::QSqrt(Double_t x) {
  return 0.5*TMath::Erfc(TMath::Sqrt(x/2.0));
}

inline Double_t  TDSPMath::Rayleigh(Double_t x, Double_t sigma) {
  return x/(sigma*sigma)*TMath::Exp(-x*x/(2*sigma*sigma));
}

inline Double_t  TDSPMath::dB(Double_t x) {
  return 10*TMath::Log10(x);
}

inline Double_t  TDSPMath::FromdB(Double_t x) {
  return TMath::Power(10,x/10);
}

inline Double_t  TDSPMath::Lorentz(Double_t x, Double_t sigma) {
  return 2.0/(TMath::Pi()*sigma*(1.0+4.0/(sigma*sigma)*x*x));
}

inline Double_t  dB(Double_t x) {
  return TDSPMath::dB(x);
}

inline Double_t  FromdB(Double_t x) {
  return TDSPMath::FromdB(x);
}

extern const Double_t Pi;

#endif
