/** TDSP *******************************************************************
                          TDSPRating.h  -  description
                             -------------------
    begin                : Fri Mar 8 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPRATING_H
#define TDSPRATING_H

#include <Rtypes.h>
#include <Riostream.h>
#include <TMath.h>
#include <TRandom.h>
#include <TError.h>

/**a class handles the rating of stochastic
   processes and interpolations etc.
   *@author Peter Jung
   */

class TDSPRating {

 protected:
  
  Double_t fRate;                          // The Rate of the Process
  UInt_t    fRelativeRate;                 // what was the last relative rate (input-rate/fRate)                 
  Double_t fRelativeRateMean;              // what is the mean relative rate
  Double_t fRelativeRateSigma;
  UInt_t   fRateIndex;                     // Counter (due to diffenent sampling rates, Representation holds)

  void     Next();                         // Next entry
  Int_t    NewRelativeRate();              // New Relative Rate
  
 public: 
  
  TDSPRating();
  virtual ~TDSPRating();

  
  void     SetRate(Double_t);                              // Set the Rate for the Stochastic Process
  Double_t GetRate() { return fRate;};                     // Get the Rate for the Stochastic Process
  Double_t GetRelativeRate() { return fRelativeRateMean;}; // Get mean of the relative ratings
  void     SetRateSigma(Double_t s) { fRelativeRateSigma = s;}; // Set the Sigma
  Double_t GetRateSigma() const { return fRelativeRateSigma;}; // Get the Sigma
  Double_t GetLastRelativeRate() { return fRelativeRate;}; // get the last relative rating
  void     ResetRating()  { fRateIndex=0;};  // Reset the Rating
  Bool_t   IsRate(); // Do we have to change the representation 
  virtual  void  PrepareRate(Double_t r); // Prepare the Rating to application on a signal sampled at a rate "r"

  ClassDef(TDSPRating,1)
  
};

inline void   TDSPRating::Next() {
  
  if (++fRateIndex>=fRelativeRate) {
    NewRelativeRate();
    fRateIndex=0;
  }
}

inline Bool_t TDSPRating::IsRate() {
  return fRateIndex == 0;
}


inline Int_t TDSPRating::NewRelativeRate() { 
  ResetRating();
  return (fRelativeRate = TMath::Max(1,TMath::Nint(gRandom->Gaus(fRelativeRateMean,fRelativeRateSigma)))); 
} 



inline void   TDSPRating::PrepareRate(Double_t rate){
  

#if defined(__OPTIMIZE__)&&defined(__GNUC__)
  // Workaround for wrong optimization in GCC 2.95.3
  //
  volatile Double_t Relative_Rate = rate/fRate;
#else
  Double_t Relative_Rate = rate/fRate;
#endif
  // The Relative Rates has been changed (this can only come from manual settings)
  //
  if (fRelativeRateMean != Relative_Rate) {
    if (Relative_Rate<1.0) 
      Error("TDSPRating::PrepareRate","Can not set relative Sampling Rates (input/response) of %f < 1.0 !!",Relative_Rate);
    else {
      
      fRelativeRateMean  = Relative_Rate;
      
      NewRelativeRate();
    }
  }
  
}


#endif
