/** TDSP *******************************************************************
                          TDSPfft.h  -  description
                             -------------------
    begin                : Mon Sep 10 2001
    copyright            : (C) 2001 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPFFT_H
#define TDSPFFT_H

#include <TDSPVector.h>
#include <TDSPMatrix.h>
#include <TDSPSignal.h>
#include <TDSPOperator.h>
#include <TDSPOutput1.h>

/**
  *@author Peter Jung
  */



class TDSPfft : public TDSPOperator  {

  friend istream &operator>>(istream&,TDSPfft&);
  friend ostream &operator<<(ostream&,TDSPfft&);
  friend istream &operator>>(istream&,TDSPfft*);
  friend ostream &operator<<(ostream&,TDSPfft*);
 

 protected:

  static    Bool_t fFFTW; ///< do we have the fftw-library ? 
  Int_t     fDim; ///< The Dimension ('r','R',1 or 'c','C',2) for the fft of incoming signals
  Double_t  fOverSampling; ///< Oversampling factor (zeropadding in ifft for >1)
  UInt_t    fCyclicPrefix; ///< The Cyclic Prefix = number of element from end inserted at the beginning 

   // my Outputs
  TDSPSignal *fFFT_Signal; ///< the fft signal-vector
  TDSPSignal *fIFFT_Signal; ///< the ifft signal-vector
    
  // low level routines from fftw-project
  //
  void *fftw_vec;         //!
  void *fftw_plan_forward;    //!
  void *fftw_plan_backward;   //!
 

  // my Slots
  TString Slot_FFT;
  TString Slot_IFFT;

  // my Signals
  TString Signal_FFT;
  TString Signal_IFFT;

  
 public:
 
  static Bool_t LoadFFTW(); ///< loads the atlas library
  static void   EnableFFTW() { fFFTW=kTRUE;}; ///< disable the use of the atlas library
  static void   DisableFFTW() { fFFTW=kFALSE;}; ///< enable the use of atlas library
  static Bool_t UsingFFTW()  { return fFFTW;}; ///< Using FFTW ?
    
  // When using tdsp in mex-files - fftw crashes in fftw_destroy_plan...

  static Bool_t _fftw_destroy_plans; 
  
  void do_fft( TComplex *, 
	       TComplex *,
	       Int_t     ndat, 
	       Int_t     stepin       = 1, 
	       Int_t     stepout      = 1,
	       Int_t     cp           = 0,
	       Double_t  oversampling = 1.0 );
 
  void do_ifft(TComplex *, 
	       TComplex *, 
	       Int_t     ndat, 
	       Int_t     stepin       = 1, 
	       Int_t     stepout      = 1,
	       Int_t     cp           = 0,
	       Double_t  oversampling = 1.0); 
  
  TDSPfft(char *name = NULL);
  ~TDSPfft();


  void        SetDim(Int_t d) { fDim=d;}; ///< Set the Dimension for fft/ifft-processing (see member fDim ...)
  Int_t       GetDim() const { return fDim;};

  TDSPSignal* GetFFTSignal() const { return fFFT_Signal;};
  TDSPSignal* GetIFFTSignal() const { return fIFFT_Signal;};
   
  // Register the slots and the signals of
  // fft-object
  
  void RegisterSignalsAndSlots() {
    
    // Add the Signals

    Signal_FFT     = AddSignal("FFT(TDSPSignal*)");
    Signal_IFFT    = AddSignal("IFFT(TDSPSignal*)");

    // Add the Slots
   
    Slot_FFT       = AddSlot("FFT(TDSPSignal*)");
    Slot_IFFT      = AddSlot("IFFT(TDSPSignal*)");
    
  }

  // Calculate the fft/ifft of vectors and matrices
  //
  void         fft(TComplex  *in, TComplex  *out, Int_t l, Int_t stepin=1, Int_t stepout=1);
  void        ifft(TComplex  *in, TComplex  *out, Int_t l, Int_t stepin=1, Int_t stepout=1);
  TDSPVector*  fft(TDSPVector*in, TDSPVector*out=NULL); ///< the fft of "in"
  TDSPMatrix*  fft(TDSPMatrix*in, Int_t dim, TDSPMatrix*out=NULL);///< the fft of the rows(dim=1,'R') or cols(dim=2,'C') of "in"

  TDSPVector* ifft(TDSPVector*in, TDSPVector*out=NULL); ///< the ifft of "in"
  TDSPMatrix* ifft(TDSPMatrix*in, Int_t dim, TDSPMatrix*out=NULL);///< the ifft of the rows(dim=1,'R') or cols(dim=2,'C') of "in"

  // Slot-Version
  //
  TDSPSignal* FFT(TDSPSignal *input);
  TDSPSignal* IFFT(TDSPSignal *input);
  
  // Get the Size of the output vector from the IFFT using
  // the current settings

  Int_t    GetIFFTLen(Int_t len) {return (Int_t)((len+fCyclicPrefix)*fOverSampling);}; // IFFT-outputlen translation
  Int_t    GetFFTLen(Int_t len)  {return (Int_t)((len/fOverSampling)-fCyclicPrefix);};  // FFT-outputlen translation

  // Set/Get the Oversampling Factor
  //
  void     SetOverSampling(Double_t o) {  fOverSampling = o; }; ///< Set the Oversampling factor
  Double_t GetOverSampling() const { return fOverSampling;}; ///< Get the Oversampling factor

  // Set/Get the Size of the Cyclic Prefix
  //
  void     SetCyclicPrefix(UInt_t c) {  fCyclicPrefix = c; }; ///< Set the length of the cyclic prefix
  UInt_t   GetCyclicPrefix() const { return fCyclicPrefix;}; ///< Get the length of the cyclic prefix

  ClassDef(TDSPfft,1)

};

// global fft-object

extern TDSPfft *gFFT; ///< the object used for the global fft/ifft-routines

// global fft-routines

TDSPVector *fft(TDSPVector *in, Int_t N=0, TDSPVector *out = NULL); ///< the N-Point fft of "in"
TDSPVector *ifft(TDSPVector *in, Int_t N=0, TDSPVector *out = NULL); ///< the N-Point ifft of "in"
TDSPMatrix *fft(TDSPMatrix *in, Int_t dim, TDSPMatrix *out = NULL); ///< the fft of the rows(dim=1,'R') or cols(dim=2,'C') of "in" 
TDSPMatrix *ifft(TDSPMatrix *in, Int_t dim, TDSPMatrix *out = NULL);///< the ifft of the rows(dim=1,'R') or cols(dim=2,'C') of "in"

#endif
