/** TDSP *******************************************************************
                          TDSPMatrix.h  -  description
                             -------------------
    begin                : Mon Feb 18 2002
    copyright            : (C) 2002 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPMATRIX_H
#define TDSPMATRIX_H

#include <TH2.h>
#include <TArrayI.h>
#include <TDSPVector.h>

/**Complex Matrices
 *@author Peter Jung
 */

///Optimizations

#define LINALG_INDEX_OPT 1
#define LINALG_MAT_ROWWISE 1


///Modes for TDSPMatrix::Apply()
//
enum eMatrixApplyMode { ///C=A*B  (A=this)
  kMatrixApplyReal,      ///Product as they are real matrices
  kMatrixApplyConjA,     ///Conjugate A before Multiplication
  kMatrixApplyConjB,     ///Conjugate B before Multiplication
  kMatrixApplyConjAB     ///Conjugate A and B before Multiplication
};

class TDSPMatrix : public TDSPVector  {

  static TH2   *fHistogram2;
  char*         fZTitle; //!/< A name for the z-axis
 protected:
  
  UInt_t fRows;   ///<Number of Rows of the Matrix
  UInt_t fCols;   ///<Number of Columns of the Matrix
  
 public: 
  
  friend istream    &operator>>(istream&,TDSPMatrix&);
  friend ostream    &operator<<(ostream&,TDSPMatrix&);
  friend istream    &operator>>(istream&,TDSPMatrix*);
  friend ostream    &operator<<(ostream&,TDSPMatrix*);

  TDSPMatrix();
  TDSPMatrix(Int_t rows, Int_t cols);

  ~TDSPMatrix();
  
  TComplex&   Swap(Int_t, Int_t, Int_t, Int_t);  ///< Swap two elements and return the first indexed (after exchanging!)
  void        SetSize(Int_t rows, Int_t cols);
  UInt_t      GetRows() const { return fRows; }; ///<Get the number rows in the matrix
  UInt_t      GetCols() const { return fCols; }; ///<Get the number columns in the matrix
  void        SetRows(UInt_t, Bool_t savedata=kFALSE); ///<Set the number of rows in the matrix
  void        SetCols(UInt_t, Bool_t savedata=kFALSE); ///<Set the number of cols in the matrix
  virtual void SetLen(Int_t num, Bool_t savedata=kFALSE); ///< Set the size
  ///for later extensions to index ranges
  
  Int_t       Clo() const { return 0;};
  Int_t       Chi() const { return fCols-1;};
  
  Int_t       Rlo() const { return 0;};
  Int_t       Rhi() const { return fRows-1;};  

  TDSPMatrix* Dup() { return Copy(NULL);};                ///<Duplicate
  TDSPMatrix* Copy(TDSPMatrix* n);                   ///<Copy the Matrix to n (=NULL - create a new Matrix )
  TDSPVector* Apply(TDSPVector*, TDSPVector *out=NULL, eMatrixApplyMode m=kMatrixApplyReal);
  TDSPVector* RightApply(TDSPVector*, TDSPVector* out =NULL); ///< A.RightApply(B,C) : C=A*B
  TDSPMatrix* RightApply(TDSPMatrix*, TDSPMatrix* out =NULL); ///< A.RightApply(B,C) : C=A*B
  TDSPVector* LeftApply( TDSPVector*, TDSPVector* out =NULL); ///< A.LeftApply(B,C) : C=B*A
  TDSPMatrix* LeftApply( TDSPMatrix*, TDSPMatrix* out =NULL); ///< A.LeftApply(B,C) : C=B*A

  TDSPVector* AdjointRightApply(TDSPVector*, TDSPVector* out =NULL); ///<A.AdjointRightApply(B,C) : C=A^* *B
  TDSPMatrix* AdjointRightApply(TDSPMatrix*, TDSPMatrix* out =NULL); ///<A.AdjointRightApply(B,C) : C=A^* *B
  TDSPVector* AdjointLeftApply( TDSPVector*, TDSPVector* out =NULL); ///<A.AdjointLeftApply(B,C) : C=B*A^*
  TDSPMatrix* AdjointLeftApply( TDSPMatrix*, TDSPMatrix* out =NULL); ///<A.AdjointLeftApply(B,C) : C=B*A^*
  TDSPVector* TransposedRightApply(TDSPVector*, TDSPVector* out =NULL); ///<A.TransposedRightApply(B,C) : C=A^T*B
  TDSPVector* TransposedLeftApply( TDSPVector*, TDSPVector* out =NULL); ///<A.TransposedRightApply(B,C) : C=A^T*B

  TDSPMatrix* Transform(TDSPMatrix* H,TDSPMatrix* result=NULL); ///<Transform the Matrix A=this : result=Transform(H)=HAH^*
  TDSPMatrix* AdjointTransform(TDSPMatrix* H,TDSPMatrix* result=NULL); ///<result=AdjointTransform(H) = H^* A H

  TDSPMatrix* Transposed(TDSPMatrix *r=NULL); ///<Creates the Transposed Matrix into r (r=NULL, create new matrix)
  void        TransposeMe(); ///< Transpose me 
  TDSPMatrix* Adjoint(TDSPMatrix *r=NULL); ///<Creates the Adjoint Matrix into r (r=NULL, create new matrix)
  TComplex    Trace(); ///< returns the trace of the matrix
  Double_t    Trace2(); ///< returns the sum of abs-squared diagonal-entries of the matrix 

  Bool_t      DiagZero(); ///<Test for 1 Zero on the Diagonal
  Bool_t      DiagZeros(); ///<Test for Zeros on the complete Diagonal
  Bool_t      IsDiag(); ///<Test for a Diagonal Matrix
  Bool_t      IsSquare() const { return fRows==fCols;}; ///<Is this a Square Matrix ?
  Bool_t      IsHermitian(); ///<Test for Hermitian Matrix

  TDSPMatrix* ReversedColumns(TDSPMatrix *r=NULL); ///< return the column-reversed matrix 
  TDSPMatrix* ReversedRows(TDSPMatrix *r=NULL); ///< return the row-reversed matrix
  void        ReverseColumns(); ///< reverse my columns
  void        ReverseRows(); ///< reverse my rows

  TDSPVector* GetDiag(TDSPVector *a=NULL); ///<Return the diagonal as a vector
  void        SetDiag(TDSPVector *a); ///<Set the diagonal to the vector
  void        SetDiag(TComplex a); ///< Set all diagonal elements to a
  void        SetNonDiag(TComplex a); ///< Set all non-diagonal elements to a
  void        AddDiag(TComplex a); ///< Add a onto the diagonal elements of the matrix
  void        AddDiag(TDSPVector* a); ///< Add the vector onto the diagonal elements of the matrix
  void        MultDiag(TComplex a); ///< Multiplies a onto the diagonal elements of the matrix
  void        MultDiag(TDSPVector* a); ///< Multiplies the vector onto the diagonal elements of the matrix

  Double_t    Redundancy() const { return ((Double_t)fRows)/((Double_t)fCols); };
  
  TDSPMatrix *SubMatrix(Int_t atRow, Int_t atCol, 
			Int_t m, Int_t n, TDSPMatrix*out=NULL); ///< extracts the mxn submatrix at (atRow,atCol)
  TDSPMatrix *TransposedSubMatrix(Int_t atRow, Int_t atCol, 
				  Int_t m, Int_t n, TDSPMatrix*out=NULL); ///< extracts the mxn submatrix at (atRow,atCol) and tranpose it
  TComplex   *MoveVec(Int_t m, Int_t n);       ///< Move the pointer in the data at (m,n)
  TDSPVector* GetColumn(Int_t c=0, TDSPVector *out=NULL); ///<extract column "c" to a vector
  TDSPVector* GetRow(Int_t r=0, TDSPVector *out=NULL);    ///<extract row "r" to a vector
  void        SetColumn(UInt_t c, TDSPVector *in, UInt_t ifrom=0, UInt_t ito=0); ///<Set column "c" from a vector content (ifrom:ito)
  void        SetRow(UInt_t r, TDSPVector *in, UInt_t ifrom=0, UInt_t ito=0);  ///<Set row "r" from a vector content (ifrom:ito)
  void        RowConvolute(TComplex* inp, TComplex *out,
			   Int_t N=0,
			   Bool_t cyclic=kTRUE ); ///< Low-Level-RowConvolution of inp to out by this maximal N elements (cyclic?)

  TDSPVector* RowConvolute(TDSPVector*inp, TDSPVector*out=NULL, 
			   Bool_t cyclic=kTRUE,
			   Bool_t zerosbefore=kTRUE); ///<Convolutes "inp" with the rows of "this" to "out" 

  TH2*        Draw(Option_t *option = "hist surf4", 
		   Double_t dx=1, Double_t xoff=0,
		   Double_t dy=1, Double_t yoff=0,
		   TH2 *h = NULL);
  void        SetZTitle(char *name); ///< Set the name of the ztitle
  void        FFTShiftMe(Int_t dim); ///< Rearrange rows(dim=1) cols(dim=2) after an fft operation

   /////////////////////////////////////////////////////
  //
  ///some operators
  //
  

#ifdef HAVE_MATLAB
  TDSPMatrix &operator=(TmxArray &m);
#endif
  
  
  TDSPMatrix  &operator=(TDSPMatrix& x) { x.Copy(this);return *this;};
  void        Print();
  void        Input();
  void        Unit(TComplex diag=1.0) { Delta(diag);}; ///<Unit/Kronecker-Delta Matrix
  void        Delta(TComplex diag=1.0);   ///<Unit/Kronecker-Delta Matrix
  void        Fourier(); ///<Init the Matrix with a Fourier-Matrix
  void        Hilbert(); ///<Init the Matrix with a Hilbert-Matrix
  
  TComplex&   Element(Int_t i, Int_t j) { return fVec[i*fCols+j];};
  TComplex&   operator()(UInt_t row, UInt_t col);

  
  // Routines based on the LU-Decomposition
  // (see LUComp.cpp)
  
  void        LUDecompose(TArrayI& Index, Int_t &d);
  void        LUBacksubst(TArrayI& Index, TDSPVector& B);
  void        LUBacksubst(TArrayI& Index, TDSPMatrix& B);
  void        LUSolveLinear(TDSPVector& B); ///<Solve the linear equation system A*x=b,  The result x overwrites b, (this) is destroyed !!
  void        LUSolveLinear(TDSPMatrix& B); ///<Solve the linear equation system (this) * X = B, X = A.Inverse * B  is returned in B,(this) is destroyed !!
  
  TDSPMatrix* Inverse(TDSPMatrix *result=NULL);  ///< Inverse Matrix
  TComplex    Det(); ///< Determinant

  // other external routines
  //(see SVD.cpp)

  TDSPMatrix* PseudoInverse(TDSPMatrix *result=NULL);
  TDSPVector* SVD(TDSPVector *S=NULL,
		  TDSPMatrix *U=NULL, 
		  TDSPMatrix *VStar=NULL); ///< does USV^* = (this) - the svd of this matrix
  ClassDef(TDSPMatrix,1)

};

#include <MatrixRoutines.h>



inline TDSPVector* TDSPMatrix::Apply(TDSPVector *inp, TDSPVector *out, eMatrixApplyMode m) {
  switch(m) {
  case kMatrixApplyReal :
    return RightApply(inp,out);
  case kMatrixApplyConjA :
    return AdjointRightApply(inp,out);
  case kMatrixApplyConjB :
    Error("Apply","error");
    return NULL;
  case  kMatrixApplyConjAB :
    Error("Apply","error");
    return NULL;
  }
  return NULL;
}

///Calculate H * (this) * H^*

inline TDSPMatrix* TDSPMatrix::Transform(TDSPMatrix* H, TDSPMatrix *result) {
  ///This we will optimize later on
  
  TDSPMatrix* b = LeftApply(H);
  result=H->AdjointLeftApply(b,result);
  delete b;
  
  return result;

}

///Calculate H^* * (this) * H

inline TDSPMatrix* TDSPMatrix::AdjointTransform(TDSPMatrix* H, TDSPMatrix *result) {
  ///This we will optimize later on
  
  TDSPMatrix* b = H->AdjointRightApply(this);
  result = H->LeftApply(b,result);
  delete b;
  return result;

}

inline TDSPMatrix* TDSPMatrix::LeftApply( TDSPMatrix *inp, TDSPMatrix *out) {
  return inp->RightApply(this,out);
}



////////////////////////////////////////////////////////////////////////////////


inline TComplex& TDSPMatrix::operator()(UInt_t row, UInt_t col) {
  if (fVec) {
    if ((row<fRows)&&(row>=0)&&(col<fCols)&&(col>=0))
      return Element(row,col);
    else 
      Error("operator()","index (%d,%d) out of bounds (%d..%d,%d..%d)",row,col,0,fRows-1,0,fCols-1);
  } else 
    Error("operator()","matrix not allocated (this->fVec=0x0)!");
  return ComplexZero;
}

inline Bool_t TDSPMatrix::DiagZero() {
  if (!IsSquare()) return kTRUE;
  for(register UInt_t i=0;i<fRows; ++i) 
    if (Element(i,i)==0.0) return kTRUE;
  return kFALSE;
}

inline Bool_t TDSPMatrix::DiagZeros() {
  Int_t min=TMath::Min(fRows,fCols);
  for(register Int_t i=0;i<min; ++i) 
    if (Element(i,i)!=0.0) return kFALSE;
  return kTRUE;
}

inline Bool_t TDSPMatrix::IsDiag() {
  for(register UInt_t i=0;i<fRows;++i)
    for(register UInt_t j=0;j<fCols;++j)
      if ((i!=j)&&(Element(i,j)!=0.0)) return kFALSE;
  return kTRUE;
}

inline TDSPVector* TDSPMatrix::GetDiag(TDSPVector *a) {
  Int_t m=TMath::Min(fRows,fCols);
  if (!a) a=new TDSPVector();
  a->SetLen(m);
  for(register Int_t i=0;i<m;++i)
    a->Element(i)=Element(i,i);
  return a;
}


inline TDSPMatrix* TDSPMatrix::Transposed(TDSPMatrix *r) {
  
  if (r==this) { 
    Error("Transposed","Can not be the same matrix - use TDSPMatrix::TransposeMe() !");
    return NULL;
  }

  if (!r) r=new TDSPMatrix(); ///cols/rows switched !!
  r->SetSize(fCols,fRows);
  for(register UInt_t i=0;i<fRows;++i)
    for(register UInt_t j=0;j<fCols;++j)
      r->Element(j,i)=Element(i,j);
  return r;

}

inline void TDSPMatrix::TransposeMe() {
  if (fCols!=fRows) 
    Error("TransposeMe","non square matrix - use TDSPMatrix::Transposed()!");
  else
    for(register UInt_t i=0;    i<fRows;++i)
      for(register UInt_t j=i+1;j<fCols;++j)
	Swap(j,i,i,j);
}

inline TDSPMatrix* TDSPMatrix::Adjoint(TDSPMatrix *r) {
  
  if (!r) r=new TDSPMatrix(fCols,fRows); ///cols/rows switched !!

  for(register UInt_t i=0;i<fRows;++i)
    for(register UInt_t j=0;j<fCols;++j) {
      r->Element(j,i).fRe = Element(i,j).fRe;
      r->Element(j,i).fIm = -Element(i,j).fIm;
    }

  return r;

}
inline Double_t Dist(TDSPMatrix *s1, TDSPMatrix *s2) {
  return s1->Dist(s2);
}
inline Double_t Dist2(TDSPMatrix *s1, TDSPMatrix *s2) {
  return s1->Dist2(s2);
}
inline Double_t Dot(TDSPMatrix *s1, TDSPMatrix *s2) {
  return s1->Dot(s2);
}
inline Double_t Dot2(TDSPMatrix *s1, TDSPMatrix *s2) {
  return s1->Dot2(s2);
}
inline Double_t Dist(TDSPMatrix& s1, TDSPMatrix& s2) {
  return s1.Dist(s2);
}
inline Double_t Dist2(TDSPMatrix& s1, TDSPMatrix& s2) {
  return s1.Dist2(s2);
}
inline Double_t Dot(TDSPMatrix& s1, TDSPMatrix& s2) {
  return s1.Dot(s2);
}
inline Double_t Dot2(TDSPMatrix& s1, TDSPMatrix& s2) {
  return s1.Dot2(s2);
}
inline Bool_t IsDiag(TDSPMatrix*s) { return s->IsDiag();}; 
inline Bool_t IsDiag(TDSPMatrix&s) { return s.IsDiag();}; 


inline TDSPMatrix*  TDSPMatrix::Copy(TDSPMatrix *output){
  
  if (!output) output = new TDSPMatrix();
  output->SetSize(fRows,fCols);

  TComplex *ovec = output->GetVec();

  for(register Int_t i=0;i<Num;++i)  
    ovec[i] = fVec[i];

  return output;  
}

//  Swaps the Elements and return a reference to the second one element which contains now the first element 
//
inline TComplex& TDSPMatrix::Swap(Int_t i, Int_t j, Int_t m, Int_t n) {
#ifdef LINALG_MAT_ROWWISE
  UInt_t    l1=i*fCols+j;
  UInt_t    l2=m*fCols+n;
#else
  UInt_t    l1=j*fRows+i;
  UInt_t    l2=n*fRows+m;
#endif
  TComplex tmp=fVec[l1];
  fVec[l1] = fVec[l2];
  fVec[l2] = tmp;
  return fVec[l2];
}

#endif
