/** TDSP *******************************************************************
                          TDSPVector.h  -  description
                             -------------------
    begin                : Thu Nov 1 2001
    copyright            : (C) 2001 by Peter Jung
    email                : jung@hhi.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TDSPVECTOR_H
#define TDSPVECTOR_H

#include <Riostream.h>
#include <TRandom.h>
#include <TObject.h>
#include <TGraph.h>
#include <TError.h>
#include <TH1.h>
#include <TH2.h>
#include <TFile.h>
#include "TComplex.h"

#ifdef WIN32
# include "../../config.h"
#else
# ifdef __MAKECINT__
#  include "../../config.h"
# else
#  include "config.h"
# endif
#endif


/**a discrete complex signal
  *@author Peter Jung
  */



#ifdef HAVE_MATLAB
class TmxArray;
#endif

extern TComplexBase _twopi;
extern TComplexBase _sqrt2;
extern TComplexBase _inv_sqrt2;

#define kTDSPVector_Conv_MatchA      0x1  ///< cut the final samples to length of a
#define kTDSPVector_Conv_MatchB      0x2  ///< cut the final samples to length of b
#define kTDSPVector_Conv_Conj        0x4  ///< Conjugate the first function
#define kTDSPVector_Conv_UseBlockLen 0x8  ///< Separat in blocklen

#define kTDSPVector_Conv_Defaults    0  ///< Defaults for convolutions

#define kStrictNone    0
#define kStrictWarning 1
#define kStrictError   2

enum  eXCorr_Options { kXCorrNone, kXCorrCoeff, kXCorrBiased, kXCorrUnBiased };

class TDSPVector : public TNamed   {

  static UInt_t fPerLine;
  static TH1   *fHistogram;

  ///The Cache-Effect
  //
  Int_t         fMinAlloc;
  Double_t      fAllocHighScale;         ///<  Alloc more memory we really need to avoid many alloc if we resize the block 
  Double_t      fAllocLowScale;          ///<  Alloc more memory we really need to avoid many alloc if we resize the block 
  

 protected:

  static UInt_t fStrict; ///< Strictmode 0=none,1=warning,2=strict
  static Bool_t fATLAS;  
  char*         fXTitle; ///< A name for the vector index
  char*         fYTitle; ///< A name for the vector content
  TComplex     *fVec;        //! Pointer to the complex signal vector
  Int_t         Num;            ///<  Len of signal vectorm
  Int_t         fPos;           ///<  mom. position for pointwise filling etc.

  TComplex     *fAllocedVec; //!
  Int_t         fAllocedNum;    ///<  
  Double_t      fSamplingRate;  ///< The Sampling Rate 
  Int_t         fZero;          ///< for absolut positioning (=pos of the zero)


 public: 


 
  static Bool_t LoadATLAS(char *name=NULL); ///< loads the atlas library
  static void   EnableATLAS() { fATLAS=kTRUE;} ///< disable the use of the atlas library
  static void   DisableATLAS() { fATLAS=kFALSE;}; ///< enable the use of atlas library
  static Bool_t UsingATLAS()  { return fATLAS;}; ///< Using ATLAS ?

  Int_t     Lo() const { return 0;};                   ///< The lower index(=0)
  Int_t     Hi() const { return Num-1;};               ///< The higher index(=len-1)
  void      Swap(Int_t, Int_t);                        ///< Swap two elements
  TComplex& Element(Int_t i) { return fVec[i];};       ///< Return i'th component of the vector
  TComplex *GetVec() { return fVec; };                 ///< Get the pointer to the data
  TComplex *MoveVec(Int_t n) { return fVec+=n;};       ///< Move the pointer in the data
  Int_t     GetLen() const { return Num; };            ///< Get the size of the data block
  Double_t  GetTime() { return Num/fSamplingRate;};    ///< Get the time (len/samplingrate)
  void      Free();                                    ///< Free all resources used by this vector
  Int_t     GetZero() { return fZero; };               ///< Get the position of the zero
  void      SetZero(Int_t z) { fZero=z; };             ///< Set the position of the zero
  void      SetVec(TComplex *s, Int_t num=-1);         ///< Set the pointer to extern data
  void      ResetSignal() { fVec = fAllocedVec;};///< Reset the pointer to intern data
  virtual void SetLen(Int_t num, Bool_t savedata=kFALSE); ///< Set the size
  void      SetMinLen(Int_t num);                      ///< Set the minimum size to alloc
  void      Set(const TComplex&);                      ///< Set all values to c
  void      Set(Double_t);                             ///< Set all values to c
  void      Set(Int_t i, TDSPVector*,Int_t p=0,Int_t len=-1); ///< Set the values starting from i from vector at p..p+len-1
  void      SetReal(Double_t);                         ///< Set all real parts to c
  void      SetImag(Double_t);                         ///< Set all imag parts to c
  void      Zeros() { Set(0.0); };
  void      Ones() { Set(1.0);};
  void      Delta(Int_t pos=0) { Zeros(); (*this)(pos)=1;  };
  void      Step(TComplex& start, TComplex& step);     ///< Init with start,start+step,start+2*step etc. 
  void      Step(TComplex& start);                     ///< Init with start,start+1,start+2 etc.    
  void      Step(Double_t  start, Double_t  step=1.0); ///< Init with start,start+step,start+2*step etc. 
  void      Step();                                    ///< Init with 0,1,2,3 ...
  void      ConjugateMe();                               ///< Conjugate all elements
  void      AdjointMe();                               ///< Adjoint the vector
  TDSPVector* Adjoint(TDSPVector *out=NULL);           ///< Adjoint the vector
  void      Append(TDSPVector*);                       ///< Append a vector to this vector
  Double_t  Dist(TDSPVector*);                         ///< Calculate the Distance
  Double_t  Dist(TDSPVector&);                         ///< Calculate the Distance
  Double_t  Dist2(TDSPVector*);                        ///< Calculate the Distance^2
  Double_t  Dist2(TDSPVector&);                         ///< Calculate the Distance^2
  TComplex  Dot(TDSPVector*, Bool_t conj=kTRUE);        ///< Calculate the Dot-Product - conjugate (this) before ?
  TComplex  Dot(TDSPVector&, Bool_t conj=kTRUE);        ///< Calculate the Dot-Product - conjugate (this) before ?
  Double_t  Dot2(TDSPVector*);                          ///< Calculate the Dot-Product^2
  Double_t  Dot2(TDSPVector&);                          ///< Calculate the Dot-Product^2
  void      RandN(Double_t mean=0, Double_t sigma=1);     ///< real parts normal dist. with (mean,sigma=1) and imag=0
  void      RandNreal(Double_t mean=0, Double_t sigma=1); ///< real parts normal dist. with (mean,sigma=1)
  void      RandNimag(Double_t mean=0, Double_t sigma=1); ///< imag parts normal dist. with (mean,sigma=1)
  void      RandNC(Double_t sigma=1);                     ///< real/imag cnormal dist. with (sigma=1) and mean=0
  void      RandNC(TComplex&, Double_t sigma=1);          ///< real/imag cnormal dist. with (mean,sigma=1)
  void      RandNC(TComplex *, Double_t sigma=1);         ///< real/imag cnormal dist. with individ. (means,sigma=1)
  void      RandNC(TComplex *, TComplex *sigmas);         ///< real/imag cnormal dist. with individ. (means,sigmas)

  void      RandNC(TDSPVector*, Double_t sigma=1);        ///< real/imag cnormal dist. with individ. (means,sigma=1)
  void      RandNC(TDSPVector&, Double_t sigma=1);        ///< real/imag cnormal dist. with individ. (means,sigma=1)
  void      RandNC(TDSPVector*, TDSPVector*sigmas);       ///< real/imag cnormal dist. with individ. (means,sigmas)
  void      RandPhase(); ///< init as random phasors (uniform phase ...)
  void      MultRandPhase(TDSPVector*in=NULL); ///< init with random phase multiplied on an input vector(if NULL use this).  
  void      RandNPhase(Double_t variance=1,Double_t rate=1); ///< init with random normal phases at a special rate
  void      MultRandNPhase(Double_t variance=1,Double_t rate=1); ///< multiply with random phase 
  void      MultRandNPhase(Double_t variance,Double_t rate,TDSPVector*); ///< init with random phase multiplied on an input vector.  
  void      Reverse();                                    ///< Reverse the order
  void      RandRayleigh(Double_t sigma=1,Double_t phase=0);///< init with real rayleigh 
  void      ModulateMe(TComplex, Double_t burstlen=0,Double_t burstoff=0, 
		       Double_t phase=0);///< Modulate the vector (multiply with a frequency)
  TDSPVector* Modulate(TComplex, Double_t burstlen=0, Double_t burstoff=0,
		       Double_t phase=0,
		       TDSPVector* = NULL);///< Modulate the vector (reset phase after each burstlen steps)

  TDSPVector* DownSampling(UInt_t M, TDSPVector *pulse=NULL, TDSPVector *output=NULL); ///< M-fold decimation filter with response "pulse"

  TDSPVector* UpSampling(  UInt_t K, TDSPVector *pulse=NULL, TDSPVector *output=NULL,
			   UInt_t iStep=1, UInt_t iOff=0); ///< K-fold interpolation filter with response "pulse" 

  TDSPVector *Alias(TDSPVector *a=NULL, Int_t NewLen=0);   ///< Creates an Alias to this vector of Length NewLen

  TComplex  Sum(TDSPVector *weights=NULL) ; ///< The Sum 
  Double_t  Sum2(TDSPVector *weights=NULL);  ///< The Sum of squares abs 
  TComplex  Prod(); ///< the product of all components
  Double_t  ProdReal(); ///< the product of all components real part
  Double_t  ProdImag(); ///< the product of all components imag part
      

  Double_t  Norm(TDSPVector *weights=NULL,Int_t p=2); ///< Calculate the weighted p-Norm (default=2-norm)
  Double_t  Norm2(TDSPVector *weights=NULL) { return Sum2(weights);};  ///< Calculate the Squared 2-Norm 

  Double_t  Energy() { return Norm2()/fSamplingRate;};    ///< Calculate the Energy (sampling-rate normalized !)
  Double_t  Power() {  return Num ? Norm2()/Num : 0;};    ///< Calculate the Power

  TComplex  Mean(TDSPVector *weights=NULL); ///< Calculate the Mean
  Double_t  Mean2(TDSPVector *weights=NULL); ///< Calculate the Mean2
  Double_t  Variance(TDSPVector *weights=NULL); ///< Calculate the Variance
  Double_t  Moment1(TDSPVector *x = NULL); ///< 1.Moment = 1/norm2(x)*sum_i Abs2(this(i))*Abs(x(i)) (x=NULL-->x=i)
  Double_t  Moment2(TDSPVector *x = NULL); ///< 2.Moment = 1/norm2(x)*sum_i Abs2(this(i))*Abs2(x(i)) (x=NULL-->x=i)
  Double_t  CMoment2(TDSPVector *x = NULL); ///< 2.CentralMoment = 1/norm2(x)*sum_i Abs2(this(i))*(Abs2(x(i))-xmean) (x=NULL-->x=i)
  Double_t  Max() { return MaxReal();};      
  Double_t  Min() { return MinReal();};      

  Double_t  MaxReal();                                    ///< Return the maximum real part
  Double_t  MinReal();                                    ///< Return the minimum real part
  Double_t  MaxImag();                                    ///< Return the maximum imag part
  Double_t  MinImag();                                    ///< Return the minimum imag part
  
  Double_t  MaxAbs2();                                     ///< Return the maximum squared abs
  Double_t  MinAbs2();                                     ///< Return the minimum squared abs

  Double_t  MaxAbs() { return TMath::Sqrt(MaxAbs2());};   ///< Return the maximum abs
  Double_t  MinAbs() { return TMath::Sqrt(MinAbs2());};   ///< Return the minimum abs
  
  TDSPVector* GetPhase(TDSPVector*resul=NULL);          ///< return the phases of the vector in a new vector
  void      Phase();                                      ///< set the vector to its phases
  void      Sqr();                                        ///< Square the vector-abs : c^* * c
  void      Abs2() { Sqr();};                             ///< Square the vector-abs : c^* * c (like sqr)
  void      Abs();                                        ///< Absolute value per index of the vector c^t*c
  void      Sqrt();                                       ///< Square Root of the Vector
  void      SqrtReal();                                   ///< Square Root only on the real part of the Vector
  void      SqrtImag();                                   ///< Square Root only on the imaginary part of the Vector
  void      Exp();                                        ///< Exp of the Vector
  void      Cos();                                        ///< Cosine of the Vector
  void      Sin();                                        ///< Sine of the Vector
  void      Log();                                        ///< nat. Log of the vector
  void      Log10();                                      ///< 10-Log of the vector
  void      Log2();                                       ///< 2-Log of the vector
  void      Normalize(Double_t norm=1.0);                 ///< Normalize
  TDSPVector *Dup() { return Copy(NULL);};                ///< Duplicate
  TDSPVector *Copy(TDSPVector* n);                        ///< Copy the Vector to n (=NULL - create a new vector )
  void      Apply(TComplex (*fcn)(TComplex&));            ///< Apply a Function on the signal vector
  void      SetSamplingRate(Double_t s) { fSamplingRate = s;};
  Double_t  GetSamplingRate() { return fSamplingRate; };
  void      SamplingRate(Double_t s) { fSamplingRate *= s;};
  void      Configure(TDSPVector*);                       ///< Configure the signal on the base on an other signal
  void      SetXTitle(char *name); ///< Set the name of the xtitle
  void      SetYTitle(char *name); ///< Set the name of the ytitle

  Int_t     OnCodeInc(TDSPVector *Alphabet);///< Increments this(seen a sequence of complex number) to the next codeword
  Int_t       OnCodeInc(TComplex *Alphabet, 
		      Int_t NumAlphabet); ///< Increments this (seen a sequence of complex number) to the next codeword
  Int_t       OnCodeSet(Int_t       value,
		      TDSPVector* al); ///< Codes the value in the Vector based on the alphabet "al"
  Int_t       OnCodeSet(Int_t     value,
		      TComplex* al,Int_t     num_al);///< Codes the value in the Vector based on the alphabet "al"
  Int_t       OnCodeGet(Int_t*      value,
		      TDSPVector* al); ///< Decodes the vector based on the alphabet "al" to *value
  Int_t       OnCodeGet(Int_t*    value,
		      TComplex* al,Int_t     num_al);///< Decodes the Vector based on the alphabet "al" to *value
    

  /// Some Pointwise operations

  TDSPVector*      DivBy(TDSPVector  *a, TDSPVector *result=NULL); 
  TDSPVector*      MultBy(TDSPVector *a, TDSPVector *result=NULL); 
  TDSPVector*      AddBy(TDSPVector  *a, TDSPVector *result=NULL);
  TDSPVector*      SubBy(TDSPVector  *a, TDSPVector *result=NULL);

  /// Some Scalar operations
    
  TDSPVector*      DivBy(TComplex  a, TDSPVector *result=NULL); 
  TDSPVector*      MultBy(TComplex a, TDSPVector *result=NULL); 
  TDSPVector*      AddBy(TComplex  a, TDSPVector *result=NULL);
  TDSPVector*      SubBy(TComplex  a, TDSPVector *result=NULL);

  TDSPVector*      Conv(TDSPVector* input,  
			TDSPVector* output = NULL,  
			Int_t       opt = kTDSPVector_Conv_Defaults,
			Int_t       b   = 0); ///< Convolution 

  void      DumpParams(ostream *os = NULL);


  void      FFTShiftMe(); ///< Rearrange after fft / before ifft (moving the zero-freq to the center)
  void      IFFTShiftMe(); ///< Rearrange after ifft / before fft (moving the zero-freq to the center)
  
  TDSPVector*  XCorr(Int_t max_lags   = 0,                    ///< AutoCorrelation
		     eXCorr_Options o = kXCorrNone) { return XCorr(this, NULL , max_lags, o);};

  TDSPVector*  XCorr(TDSPVector *i2,                          ///< Crosscorrelation 
		     Int_t max_lags   = 0,
		     eXCorr_Options o = kXCorrNone) { return XCorr(i2,   NULL , max_lags, o); };

  TDSPVector*  XCorr(TDSPVector *i2,                          ///< Crosscorrelation
		     TDSPVector *result, 
		     Int_t max_lags   = 0,
		     eXCorr_Options o = kXCorrNone);

  TGraph*   Graph(Option_t* = "complex", TGraph *in=NULL, Double_t dx=1.0, Double_t xoff=0.0);
  virtual void     Draw(Option_t* = "ac", Double_t dx=1.0, Double_t xoff=0.0); //*MENU*
  TH1*      FillAbs(TH1 *h = NULL); //*MENU*
  TH1*      FillAbs2(TH1 *h = NULL); //*MENU*
  TH1*      FillPhase(TH1 *h = NULL); //*MENU*
  TH1*      FillReal(TH1 *h = NULL); //*MENU*
  TH1*      FillImag(TH1 *h = NULL); //*MENU*
  TH2*      Fill(TH2 *h = NULL); //*MENU*
  virtual void      Print(Option_t*option="");
  void      Input();

  /////////////////////////////////////////////////////
  //
  ///some operators
  //
  

#ifdef HAVE_MATLAB
  TDSPVector &operator=(TmxArray &m);
#endif

  TDSPVector &operator=(const Double_t x)  { Set(x); return *this;};
  TDSPVector &operator=(TDSPVector& x) { x.Copy(this);return *this;};
  TDSPVector &operator=(const TComplex& c) { Set(c); return *this;};
  TDSPVector &operator&=(const Double_t&); ///Add a value into the Vector and move the cursor

  TDSPVector &operator/=(const TComplex& d);
  TDSPVector &operator/=(const Double_t& d);
  TDSPVector &operator/=(const Int_t& d);
  TDSPVector &operator*=(const TComplex& d);
  TDSPVector &operator*=(const Double_t& d);
  TDSPVector &operator*=(const Int_t& d);
  TDSPVector &operator+=(const TComplex& d);
  TDSPVector &operator+=(const Double_t& d);
  TDSPVector &operator+=(const Int_t& d);
  TDSPVector &operator-=(const TComplex& d);
  TDSPVector &operator-=(const Double_t& d);
  TDSPVector &operator-=(const Int_t& d);

  TDSPVector &operator+=(const TDSPVector&s);
  TDSPVector &operator-=(const TDSPVector&s);
  TDSPVector &operator*=(const TDSPVector&s);
  TComplex   &operator()(Int_t index);
  TComplex   &operator[](Int_t index);

  static TDSPVector* Load(TFile*    f, char *key); ///< Load a Vector from a file/key
  static TDSPVector* Load(char *fname, char *key); ///< Load a Vector from a file/key

  void   Save(TFile*            f, char *key); ///< Save a Vector into a file/key
  void   Save(char *fname        , char *key, char *opt="RECREATE"); ///< Save a Vector into a file/key


  TDSPVector(Int_t len=0);
  virtual ~TDSPVector();

  static void ValuesPerLine(UInt_t p) { fPerLine = p;}; ///How many Values to Print per line (using cout...)
 
  friend Bool_t      operator==(TDSPVector &, TDSPVector &);
  friend istream    &operator>>(istream&,TDSPVector&);
  friend ostream    &operator<<(ostream&,TDSPVector&);
  friend istream    &operator>>(istream&,TDSPVector*);
  friend ostream    &operator<<(ostream&,TDSPVector*);

  friend Double_t      filternorm(TDSPVector *fir,
				  TDSPVector *iir = NULL,
				  UInt_t      p   = 2);

  friend TDSPVector*   filter(TDSPVector *in,
			      TDSPVector *out,
			      TDSPVector *fir,
			      TDSPVector *iir=NULL);
  
  friend void     mult(TDSPVector *a,TDSPVector *b,
		       TDSPVector *c);


  friend void    econv(TDSPVector *a,
		       TDSPVector *b,
		       TDSPVector *c,
		       Int_t offset = 0,
		       Int_t step   = 1,
		       Int_t opt =kTDSPVector_Conv_Defaults );
		       
  ClassDef(TDSPVector,1)
    
};
  

inline TDSPVector& TDSPVector::operator&=(const Double_t& d) {
  fVec[fPos++] = d;
  return *this;
}

inline TDSPVector& TDSPVector::operator/=(const Double_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]/=d;
  return *this;
}

inline TDSPVector& TDSPVector::operator/=(const Int_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]/=d;
  return *this;
  
}

inline TDSPVector& TDSPVector::operator/=(const TComplex& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]/=d;
  return *this;
}
inline TDSPVector& TDSPVector::operator*=(const Double_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]*=d;
  return *this;
}

inline TDSPVector& TDSPVector::operator*=(const Int_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]*=d;
  return *this;
  
}

inline TDSPVector& TDSPVector::operator*=(const TComplex& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]*=d;
 
  return *this;
}

inline TDSPVector& TDSPVector::operator+=(const Double_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]+=d;
  return *this;
}

inline TDSPVector& TDSPVector::operator+=(const Int_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]+=d;
  return *this;
  
}

inline TDSPVector& TDSPVector::operator+=(const TComplex& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]+=d;
  return *this;
}


inline TDSPVector& TDSPVector::operator-=(const Double_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]-=d;
  return *this;
}

inline TDSPVector& TDSPVector::operator-=(const Int_t& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]-=d;
  return *this;
  
}

inline TDSPVector& TDSPVector::operator-=(const TComplex& d) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]-=d;
 
  return *this;
}




inline void TDSPVector::SetVec(TComplex *s,Int_t l ) {
  if (l!=-1) Num  = l;
  fVec = s;
}

inline void TDSPVector::Set(const TComplex& c) {
  for(register Int_t i=0;i<Num;++i) fVec[i] = c;
}

inline void TDSPVector::Set(Double_t d) {
  for(register Int_t i=0;i<Num;++i) { fVec[i].fRe = d; fVec[i].fIm = 0; };
}

inline void TDSPVector::SetReal(Double_t d) {
  for(register Int_t i=0;i<Num;++i) fVec[i].fRe = d;
}

inline void TDSPVector::SetImag(Double_t d) {
  for(register Int_t i=0;i<Num;++i) fVec[i].fIm = d;
}

inline void TDSPVector::AdjointMe() {
  for(register Int_t i=0;i<Num;++i) fVec[i].fIm *= -1.0;
}

inline TDSPVector* TDSPVector::Adjoint(TDSPVector *out) {
  if (!out) out = new TDSPVector();
  out->SetLen(Num);
  for(register Int_t i=0;i<Num;++i) {
    out->fVec[i].fRe=fVec[i].fRe;
    out->fVec[i].fIm=fVec[i].fIm * -1.0;
  }
  out->Configure(this);
  return out;
}

inline void TDSPVector::Reverse() {
  Int_t  n=Num/2;
  Int_t n1=Num-1;
  TComplex tmp;
  for(register Int_t i=0;i<n;++i) {
    tmp=fVec[n1-i];
    fVec[n1-i] = fVec[i];
    fVec[i]=tmp;
  }
}

inline void TDSPVector::RandNreal(Double_t mean, Double_t sigma) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i].fRe = gRandom->Gaus(mean,sigma);
}

inline void TDSPVector::RandNimag(Double_t mean, Double_t sigma) {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i].fIm = gRandom->Gaus(mean,sigma);
}

inline void TDSPVector::RandN(Double_t mean, Double_t sigma) {
  for(register Int_t i=0;i<Num;++i) { 
    fVec[i].fRe = gRandom->Gaus(mean,sigma);
    fVec[i].fIm = 0;
  }
}


inline void TDSPVector::RandNC(TDSPVector& means, Double_t sigma) {
  RandNC(&means, sigma);
}



inline void TDSPVector::Step(TComplex& start, TComplex& step) {
  TComplex s = start;
  for(register Int_t i=0;i<Num;++i) { 
    fVec[i] = s;
    s+=step;
  }  
}

inline void TDSPVector::Step(TComplex& start) {
  TComplex s = start;
  for(register Int_t i=0;i<Num;++i) { 
    fVec[i] = s;
    s+=1;
  }  
}

inline void TDSPVector::Step(Double_t start, Double_t step) {
  for(register Int_t i=0;i<Num;++i) { 
    fVec[i] = start;
    start+=step;
  }  
}

inline void TDSPVector::Step() {
  for(register Int_t i=0;i<Num;++i)
    fVec[i] = i; 
}

inline void TDSPVector::RandRayleigh(Double_t sigma, Double_t phase) {
  TComplexBase tmp1,tmp2;
  for(register Int_t i=0;i<Num;++i) { 
    tmp1=gRandom->Gaus(0,sigma);
    tmp2=gRandom->Gaus(0,sigma);
    tmp1=TMath::Sqrt(tmp1*tmp1+tmp2*tmp2);
   
    fVec[i].fRe = TMath::Cos(phase)*tmp1;
    fVec[i].fIm = TMath::Sin(phase)*tmp1;
  }
}

inline void TDSPVector::Sqr() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= fVec[i].fRe*fVec[i].fRe + fVec[i].fIm*fVec[i].fIm;
}


inline void TDSPVector::Abs() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Abs(fVec[i]);
}

inline void TDSPVector::Sqrt() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Sqrt(fVec[i]);
}

inline void TDSPVector::Exp() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Exp(fVec[i]);
}
inline void TDSPVector::Log() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Log(fVec[i]);
}
inline void TDSPVector::Log10() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Log10(fVec[i]);
}
inline void TDSPVector::Log2() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Log2(fVec[i]);
}
inline void TDSPVector::Cos() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Cos(fVec[i]);
}
inline void TDSPVector::Phase() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= fVec[i].Phase();
}
inline void TDSPVector::Sin() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i]= ::Sin(fVec[i]);
}

inline void TDSPVector::SqrtReal() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i].fRe = TMath::Sqrt(fVec[i].fRe);
}

inline void TDSPVector::SqrtImag() {
  for(register Int_t i=0;i<Num;++i) 
    fVec[i].fIm = TMath::Sqrt(fVec[i].fIm);
}

inline void TDSPVector::Normalize(Double_t norm) {
  Double_t f = norm/Norm();
  (*this)*=f;
}


inline Double_t TDSPVector::Dot2(TDSPVector *s) {
  Double_t dot2=0;
  if (s->GetSamplingRate()!=GetSamplingRate()) 
    Warning("Dot2","Sampling rates differ (%f != %f) !",GetSamplingRate(),s->GetSamplingRate());
  for(register Int_t i=0;i<Num;++i) {
    dot2  += fVec[i].fRe*s->fVec[i].fRe + fVec[i].fIm*s->fVec[i].fIm;
  }
  
  return dot2;
}
inline Double_t TDSPVector::Dist2(TDSPVector& s) {
  return Dist2(&s);
}
inline Double_t TDSPVector::Dist(TDSPVector* s) {
  return TMath::Sqrt(Dist2(s));
}
inline Double_t TDSPVector::Dist(TDSPVector& s) {
  return Dist(&s);
}

inline Double_t TDSPVector::Dot2(TDSPVector& s) {
  return Dot2(&s);
}
inline TComplex TDSPVector::Dot(TDSPVector* s, Bool_t conj) {
  if (s->GetSamplingRate()!=GetSamplingRate()) 
    Warning("Dot","Sampling rates differ (%f != %f) !",GetSamplingRate(),s->GetSamplingRate());
  
  TComplex dot=0;
  if (conj) 
    for(register Int_t i=0;i<Num;++i) 
      dot  += fVec[i] % s->fVec[i];
  else 
    for(register Int_t i=0;i<Num;++i) 
      dot  += fVec[i] * s->fVec[i];
  return dot;
}
inline TComplex TDSPVector::Dot(TDSPVector& s, Bool_t conj) {
  return Dot(&s,conj);
}

inline TComplex& TDSPVector::operator()(Int_t index) {
  if (fVec) {
    if ((index<Num)&&(index>=0))
      return fVec[index];
    else 
      Error("operator()","index %d out of bounds (%d..%d)",index,0,Num-1);
  } else 
    Error("operator()","vector not allocated (this->fVec=0x0)!");
  return ComplexZero;
}

inline TComplex& TDSPVector::operator[](Int_t index) {
  return fVec[index];
}

inline Bool_t operator ==(TDSPVector &z1, TDSPVector &z2) {
  if (z1.GetSamplingRate()!=z2.GetSamplingRate()) {
    Warning("Dot2","Sampling rates differ (%f != %f) - vectors are NOT equal !",z1.GetSamplingRate(),z2.GetSamplingRate());
    return kFALSE;
  }
  Int_t     len1 = z1.GetLen();
  Int_t     len2 = z2.GetLen();
  
  TComplex *vec1 = z1.GetVec();
  TComplex *vec2 = z2.GetVec();
  
  if (len2<len1) {
    register Int_t tmp = len1;
    len1=len2;
    len2=tmp;
    vec1 = z2.GetVec();
    vec2 = z1.GetVec();
  } else {
    vec1 = z1.GetVec();
    vec2 = z2.GetVec();
    
  }

  register Int_t i;
  for(i=0;i<len1;++i) 
    if (vec1[i]!=vec2[i]) return kFALSE;  
  
  ///if the vector have not equal length the 
  ///rest of the larger one has to be zero

  for(;i<len2;++i)
    if (vec2[i].fRe || vec2[i].fIm) return kFALSE;
  
  return kTRUE;
}

inline void  TDSPVector::Apply(TComplex (*fcn)(TComplex&)){
  
  for(register Int_t i=0;i<Num;++i)  
    fVec[i] = (*fcn)(fVec[i]);

}

inline TDSPVector*  TDSPVector::Copy(TDSPVector *output){
  
  if (!output) output = new TDSPVector();

  output->SetLen(Num);
  TComplex *ovec = output->GetVec();

  for(register Int_t i=0;i<Num;++i)  
    ovec[i] = fVec[i];

  output->Configure(this);

  return output;

}

inline TDSPVector* TDSPVector::Alias(TDSPVector *a, Int_t NewNum) {
  if (!a) a = new TDSPVector();
  a->Free();
  a->Configure(this);
  if (!NewNum) NewNum=Num;
  a->SetVec(fVec,NewNum);
  return a;
}

inline void TDSPVector::Configure(TDSPVector *s) {
  SetSamplingRate(s->GetSamplingRate());
}


inline Double_t TDSPVector::MaxReal() {
  Double_t max = fVec->fRe;
  Double_t m   = 0;
  for(register Int_t i=1;i<Num;++i) if ((m=fVec[i].fRe)>max) max=m;
  return max;
}

inline Double_t TDSPVector::MinReal() {
  Double_t min = fVec->fRe;
  Double_t m   = 0;
  for(register Int_t i=1;i<Num;++i) if ((m=fVec[i].fRe)<min) min=m;
  return min;
}
inline Double_t TDSPVector::MaxImag() {
  Double_t max = fVec->fIm;
  Double_t m   = 0;
  for(register Int_t i=1;i<Num;++i) if ((m=fVec[i].fIm)>max) max=m;
  return max;
}

inline Double_t TDSPVector::MinImag() {
  Double_t min = fVec->fIm;
  Double_t m   = 0;
  for(register Int_t i=1;i<Num;++i) if ((m=fVec[i].fIm)<min) min=m;
  return min;
}
inline Double_t TDSPVector::MaxAbs2() {
  Double_t max = fVec->fIm*fVec->fIm+fVec->fRe*fVec->fRe;
  Double_t m   = 0;
  for(register Int_t i=1;i<Num;++i) 
    if ((m=fVec[i].fIm*fVec[i].fIm+fVec[i].fRe*fVec[i].fRe)>max) max=m;
  return max;
}

inline Double_t TDSPVector::MinAbs2() {
  Double_t min = fVec->fIm*fVec->fIm+fVec->fRe*fVec->fRe;
  Double_t m   = 0;
  for(register Int_t i=1;i<Num;++i) 
    if ((m=fVec[i].fIm*fVec[i].fIm+fVec[i].fRe*fVec[i].fRe)<min) min=m;
  return min;
}

inline void TDSPVector::Swap(Int_t i, Int_t j) {
  TComplex tmp=fVec[i];
  fVec[i] = fVec[j];
  fVec[j] = tmp;
}

inline TComplex TDSPVector::Mean(TDSPVector *weights) {
  if (weights) return Sum(weights)/weights->Norm2();
  return Sum(weights)/Num;
}
inline Double_t TDSPVector::Mean2(TDSPVector *weights) {
  if (weights) return Sum2(weights)/weights->Norm2();
  return Sum2(weights)/Num;
}

#include <VectorRoutines.h>

#endif
