/* Simulation of a Rayleigh flat fading MIMO channel
   06-03-2002
   Oliver Henkel (henkel@hhi.de)
*/

#ifndef __CINT__

#include <iostream.h>
#include <fstream.h>
#include <TGraph.h>
#include <TGraphErrors.h>
#include <TMultiGraph.h>
#include <TPad.h>
#include <TCanvas.h>
#include <TBenchmark.h>
#include <TLegend.h>
#include <string.h>
#include <TError.h>
#include <TDSPMatrix.h>

#endif

void simulation() {
  
  gROOT->Reset();


  //
  //
  // ------------------------------------------------------------------------  
  // Main variables
  // ------------------------------------------------------------------------
  //
  //

  const Int_t Mt = 2;  // Number of transmitter antennas
  const Int_t Mr = 2;  // Number of receiver antennas

  const Int_t MaxAllowedCodeLength=256;
  const Int_t MaxNumSNRsteps=100;
  Double_t   SNRstart = 0.0;
  Double_t   SNRstop = 12.0;
  Double_t   SNRstep = 1.0;

  // ToleranceLevel determines the allowed uncertainty in the measuring
  // of the SymbolErrors, hence it determines NumIterations
  const Double_t ToleranceLevel = 0.05;

  char *CodeFileName = "Alamouti16";    
  char *CompareFileName = "";   // Left blank, if there is none




  //
  //
  // ------------------------------------------------------------------------  
  // Reading Code
  // ------------------------------------------------------------------------
  //
  //

  Int_t ActualCodeLength;
  TDSPMatrix Symbols[MaxAllowedCodeLength];
  char CodeFilePath[100];
  sprintf(CodeFilePath,"data/%s",CodeFileName);

  ifstream CodeFile(CodeFilePath,ios::in);  
  if (CodeFile) {
    Int_t i=0;
    while(!CodeFile.eof()) {
      TDSPMatrix& Symbol=Symbols[i];
      Symbol.SetSize(Mt,Mt);   
      if (i<MaxAllowedCodeLength) {
	CodeFile >> Symbol(0,0).fRe >> Symbol(0,1).fRe 
                 >> Symbol(0,0).fIm >> Symbol(0,1).fIm;
	CodeFile >> Symbol(1,0).fRe >> Symbol(1,1).fRe 
                 >> Symbol(1,0).fIm >> Symbol(1,1).fIm;
      } else {
	Error("simulation",
	      "Maximum Number %d of Symbols exceeded! Skipping the rest...",i);
	break;
      }
      if (!CodeFile.eof()) {
	// cout << "---" << i << "--------------------------------" << endl;
	// cout << Symbols[i] << endl;
	i++;
      }
    }
    ActualCodeLength=i;
  } else {
    Error("simulation","Error reading the code file %s !",CodeFileName);
    return;
  }
  Info("simulation","Got code %s of length %d ",CodeFilePath,ActualCodeLength);

  //
  //
  // ------------------------------------------------------------------------  
  // Reading previously simulated results
  // ------------------------------------------------------------------------
  //
  //

  Double_t CompareSNR[MaxNumSNRsteps], CompareSER[MaxNumSNRsteps],
           CompareUncertainty[MaxNumSNRsteps];
  Int_t NumCompareSER=0;

  char CompareFilePath[100];
  sprintf(CompareFilePath,"out/%s",CompareFileName);

  if (CompareFileName != "") {
    ifstream CompareFile(CompareFilePath,ios::in);  
    if (CompareFile) {
      while(!CompareFile.eof()) {
	CompareFile >> CompareSNR[NumCompareSER] 
                    >> CompareSER[NumCompareSER]
	            >> CompareUncertainty[NumCompareSER];
	NumCompareSER++;
      }
      NumCompareSER--;
    }
    else {
      Error("simulation","Error reading the CompareFile %s !",CompareFileName);
      return;
    }
    Info("simulation","Got CompareData from %s",CompareFilePath);
  }
  else
    Info("simulation","No CompareData file");
  

  //
  //
  // -----------------------------------------------------------------------
  // Set up Graphic Device
  // -----------------------------------------------------------------------
  //
  //

  enum farbe{black = 1, red = 2, green = 3, blue = 4};
  TCanvas *win = new TCanvas("Simulation","Simulation",850,600);
  TGraph *grCompare = new TGraph();
  TGraph *gr = new TGraph();
  // Set properties
  win->SetGridx();
  win->SetGridy();
  win->SetLogy();
  gr->SetMarkerStyle(4);
  gr->SetMarkerSize(0.5);
  gr->SetLineColor(red);
  if (NumCompareSER) {     // compare-data exist
    grCompare->SetMarkerStyle(4);
    grCompare->SetMarkerSize(0.5);
    grCompare->SetLineColor(blue);
    for(Int_t i=0;i<NumCompareSER;i++) {
      grCompare->SetPoint(i,CompareSNR[i],CompareSER[i]);
    }
    grCompare->Draw("alp");    
 }



  //
  //
  // -----------------------------------------------------------------------
  // Simulate data transfer trough a MIMO-Rayleigh channel
  // -----------------------------------------------------------------------
  //
  //


  //
  // Declarations
  const Int_t CodeLength=ActualCodeLength;
  TDSPMatrix Channel(Mr,Mt);  
  TDSPMatrix ReceivedSymbol;
  TDSPMatrix TestSymbols[CodeLength];
  Double_t   Distances[CodeLength];
  Int_t      DetectedSymbol=0;

  // Counts the number of channel uses
  UInt_t NumIterations;
  // #symbols transmitted through each fixed channel
  const Int_t LengthDataString = 2*CodeLength;
  cout << " Length of Datastring (#fixed random channel uses) = " 
       << LengthDataString << endl;

  // keep track of the results
  Double_t SNR[MaxNumSNRsteps], SER[MaxNumSNRsteps], 
           Uncertainty[MaxNumSNRsteps];
  char OutFilePath[100];
  char *OutFileName = CodeFileName;
  sprintf(OutFilePath,"out/%s",OutFileName);
  ofstream OutFile(OutFilePath,ios::out);   // OutputFile
  
  //
  // Start Timer
  gBenchmark->Reset();
  gBenchmark->Start("time");  

  //
  // Calculate the necessary SymbolErrors to achieve given ToleranceLevel
  // (Assumes some Poisson statistics in the occurence of SymbolErrors)
  Int_t DesiredSymbolErrors = Int_t(pow((log10(TMath::E())/ToleranceLevel),2));
  cout << " ToleranceLevel " << ToleranceLevel 
       << " is achieved by " << DesiredSymbolErrors
       << " errors " << endl;

  //
  // Start Simulation
  cout << " Simulated SNR range = " 
       << SNRstart << ":"  << SNRstep << ":" << SNRstop 
       << endl;

  Int_t iSER=0;   // counter for the attained SNR values
  for(Double_t snr=SNRstart;snr<=SNRstop;snr+=SNRstep) {
    Int_t SymbolErrors=0;  // counts the symbol errors

    //
    // Start Iteration (until SymbolErrros>DesiredSymbolErrors)
    for(NumIterations=1;;NumIterations++) {  
      if (NumIterations%(LengthDataString)==0) {
	// Create new random channel matrix only every period 
	// of length LengthDataString
	Channel.RandNC(1.0); 
        // Send the alphabet through, creating TestSymbols for the decoder
        for(Int_t j = 0; j < CodeLength; j++) {
	  Channel.RightApply(&Symbols[j],&TestSymbols[j]);
	} // j
      } // if (NumIterations%(LengthDataString)==0)
      //
      // Select a random symbol from the code
      Int_t iSymbol = (Int_t)(gRandom->Rndm()*CodeLength);
      // Create the Noise
      ReceivedSymbol.RandNC( sqrt( pow(10.0,-snr/10.0) ) );
      // Add the Symbol onto the Noise
      ReceivedSymbol+=TestSymbols[iSymbol];

      //
      // Calculate the distances 
      Double_t Min=9999999;
      for(Int_t j = 0; j < CodeLength; ++j) {
	Distances[j] = Dist(ReceivedSymbol,TestSymbols[j]);
	if (Distances[j]<Min) {
	  Min = Distances[j];
	  DetectedSymbol = j;
	} // if
      } // j

      //
      // Incorrect Detection ?
      if (DetectedSymbol!=iSymbol) {
	++SymbolErrors;
	if (SymbolErrors>DesiredSymbolErrors) break;
        cout << "  SNR = " << snr 
             << "  SER = " << Double_t(SymbolErrors)/NumIterations 
             << "  #SER = " << SymbolErrors 
             << "  Uncertainty = " << TMath::Sqrt(Double_t(SymbolErrors))/NumIterations 
             << "  #Iterations/Datastring " << NumIterations/LengthDataString 
             << "   " << "\r" << flush;  
	win->Update();
      }

    } // for (NumIterations)

    //
    // DesiredSymbolError's detected
    if (SymbolErrors) {
      SNR[iSER] = snr;
      SER[iSER]  = Double_t(SymbolErrors)/NumIterations;
      Uncertainty[iSER] = TMath::Sqrt(Double_t(SymbolErrors))/NumIterations;
      // Write result into OutputFile
      OutFile  << SNR[iSER] << "   " 
               << SER[iSER] << "   " 
               << Uncertainty[iSER] << endl;
      OutFile.flush();
      // Print Information on the screen
      cout << "  SNR = " << SNR[iSER] << "  SER = " << SER[iSER] 
           << "  #SER = " << SymbolErrors << "  Uncertainty = "
           << Uncertainty[iSER] << "  #Iterations/Datastring " 
           << NumIterations/LengthDataString << endl;  
      //
      // Update the graph
      if (NumCompareSER) {   // compare-data exist
	grCompare->Draw("alp");    
	grCompare->GetXaxis()->SetTitle("SNR[db]");
	grCompare->GetYaxis()->SetTitle("SER");
	grCompare->GetXaxis()->CenterTitle();
	grCompare->GetYaxis()->CenterTitle();
	gr->SetPoint(iSER,snr,SER[iSER]);
	//gr->SetPointError(iSER,0.0,Uncertainty[iSER]);
	gr->Draw("lp");
      }
      else {
	gr->SetPoint(iSER,snr,SER[iSER]);
	//gr->SetPointError(iSER,0.0,Uncertainty[iSER]);
	if (iSER) {  // 1. gr must contain at least one point to have 
	             //    axes defiened
                     // 2. SetPoint MUST be called before Get?axis!
	  gr->GetXaxis()->SetTitle("SNR[db]");
	  gr->GetYaxis()->SetTitle("SER");
	  gr->GetXaxis()->CenterTitle();
	  gr->GetYaxis()->CenterTitle();
	}
	gr->Draw("ALP");
      }
      win->Update();
      iSER++;
    } // if (SymbolErrors)
  } // snr


  //  
  // Stop Timer
  gBenchmark->Stop("time");
  gBenchmark->Show("time");  // show elapsed time


  //
  //
  // -----------------------------------------------------------------------
  // Draw final graph and legend 
  // -----------------------------------------------------------------------
  //
  //

  //
  // Draw final graph (with axes adjusted if necessary)
  TGraphErrors *grerr = new TGraphErrors();
  TGraphErrors *grerrCompare = new TGraphErrors();
  TMultiGraph *mg = new TMultiGraph();
  // Set properties
  grerr->SetMarkerStyle(4);
  grerr->SetMarkerSize(0.5);
  grerr->SetLineColor(red);
  for(Int_t i=0;i<iSER;i++) {
    grerr->SetPoint(i,SNR[i],SER[i]);
    grerr->SetPointError(i,0.0,Uncertainty[i]);
  }
  win->Clear();
  if (NumCompareSER) {
    grerrCompare->SetMarkerStyle(4);
    grerrCompare->SetMarkerSize(0.5);
    grerrCompare->SetLineColor(blue);
    for(Int_t i=0;i<NumCompareSER;i++) {
      grerrCompare->SetPoint(i,CompareSNR[i],CompareSER[i]);
      grerrCompare->SetPointError(i,0.0,CompareUncertainty[i]);
    }
    mg->Add(grerrCompare);
    mg->Add(grerr);
    mg->Draw("alp");
    mg->GetXaxis()->SetTitle("SNR[db]");
    mg->GetYaxis()->SetTitle("SER");
    mg->GetXaxis()->CenterTitle();
    mg->GetYaxis()->CenterTitle();
  }
  else {
  grerr->Draw("alp");    
  grerr->GetXaxis()->SetTitle("SNR[db]");
  grerr->GetYaxis()->SetTitle("SER");
  grerr->GetXaxis()->CenterTitle();
  grerr->GetYaxis()->CenterTitle();
  }
  win->Update();

  //
  // Draw legend
  char headertext[100], entry[100], compareentry[100];
  sprintf(headertext,"MIMO %dx%d ", Mt,Mr);
  sprintf(entry,"%s ",CodeFileName);
  sprintf(compareentry,"%s ",CompareFileName);
  TLegend *leg = new TLegend(0.5,0.75,0.875,0.875,headertext);
  leg->AddEntry(gr,entry,"pl");
  if (NumCompareSER) 
    leg->AddEntry(grCompare,compareentry,"pl");
  leg->Draw();
}
